import pygmsh
import numpy as np

__pygmsh6__ = False
if hasattr(pygmsh, "built_in"): __pygmsh6__ = True

# ------------------------------------- #
def unitline(h):
    if __pygmsh6__:
        geom = pygmsh.built_in.Geometry()
        p0 = geom.add_point([0, 0, 0], lcar=h)
        p1 = geom.add_point([1, 0, 0], lcar=h)
        p = geom.add_line(p0, p1)
        geom.add_physical(p0, label=10000)
        geom.add_physical(p1, label=10001)
        geom.add_physical(p, label=1000)
    else:
        with pygmsh.geo.Geometry() as geom:
            p0 = geom.add_point([0, 0, 0], mesh_size=h)
            p1 = geom.add_point([1, 0, 0], mesh_size=h)
            p = geom.add_line(p0, p1)
            geom.add_physical(p0, label="10000")
            geom.add_physical(p1, label="10001")
            geom.add_physical(p, label="1000")
    return geom

# ------------------------------------- #
def unitsquare(h):
    a=1
    if __pygmsh6__:
        geom = pygmsh.built_in.Geometry()
        p = geom.add_rectangle(xmin=-a, xmax=a, ymin=-a, ymax=a, z=0, lcar=h)
        geom.add_physical(p.surface, label=100)
        for i in range(4): geom.add_physical(p.line_loop.lines[i], label=1000 + i)
    else:
        with pygmsh.geo.Geometry() as geom:
            p = geom.add_rectangle(xmin=-a, xmax=a, ymin=-a, ymax=a, z=0, mesh_size=h)
            geom.add_physical(p.surface, label="100")
            for i in range(len(p.lines)): geom.add_physical(p.lines[i], label=f"{1000 + i}")
    return geom


# ------------------------------------- #
def unitcube(h):
    if __pygmsh6__:
        geom = pygmsh.built_in.Geometry()
        x, y, z = [-1, 1], [-1, 1], [-1, 1]
        p = geom.add_rectangle(xmin=x[0], xmax=x[1], ymin=y[0], ymax=y[1], z=z[0], lcar=h)
        geom.add_physical(p.surface, label=100)
        axis = [0, 0, z[1] - z[0]]
        top, vol, ext = geom.extrude(p.surface, axis)
        geom.add_physical(top, label=105)
        geom.add_physical(ext[0], label=101)
        geom.add_physical(ext[1], label=102)
        geom.add_physical(ext[2], label=103)
        geom.add_physical(ext[3], label=104)
        geom.add_physical(vol, label=10)
    else:
        with pygmsh.geo.Geometry() as geom:
            x, y, z = [-1, 1], [-1, 1], [-1, 1]
            p = geom.add_rectangle(xmin=x[0], xmax=x[1], ymin=y[0], ymax=y[1], z=z[0], mesh_size=h)
            geom.add_physical(p.surface, label="100")
            axis = [0, 0, z[1] - z[0]]
            top, vol, ext = geom.extrude(p.surface, axis)
            geom.add_physical(top, label="105")
            geom.add_physical(ext[0], label="101")
            geom.add_physical(ext[1], label="102")
            geom.add_physical(ext[2], label="103")
            geom.add_physical(ext[3], label="104")
            geom.add_physical(vol, label="10")
    return geom


# ------------------------------------- #
def backwardfacingstep(h=1.):
    geom = pygmsh.built_in.Geometry()
    X = []
    X.append([-1.0,  1.0])
    X.append([-1.0,  0.0])
    X.append([ 0.0,  0.0])
    X.append([ 0.0, -1.0])
    X.append([ 3.0, -1.0])
    X.append([ 3.0,  1.0])
    p = geom.add_polygon(X=np.insert(np.array(X), 2, 0, axis=1), lcar=h)
    geom.add_physical(p.surface, label=100)
    ll = p.line_loop
    for i in range(len(ll.lines)): geom.add_physical(ll.lines[i], label=1000+i)
    return geom

# ------------------------------------- #
def backwardfacingstep3d(h):
    geom = pygmsh.built_in.Geometry()
    X = []
    X.append([-1.0,  1.0])
    X.append([-1.0,  0.0])
    X.append([ 0.0,  0.0])
    X.append([ 0.0, -1.0])
    X.append([ 3.0, -1.0])
    X.append([ 3.0,  1.0])
    p = geom.add_polygon(X=np.insert(np.array(X), 2, -1.0, axis=1), lcar=h)
    geom.add_physical(p.surface, label=100)
    axis = [0, 0, 2]
    top, vol, ext = geom.extrude(p.surface, axis)
    next = len(ext)
    geom.add_physical(top, label=101+next)
    for i in range(next):
        geom.add_physical(ext[i], label=101+i)
    geom.add_physical(vol, label=10)
    return geom

# ------------------------------------- #
def equilateral(h):
    geom = pygmsh.built_in.Geometry()
    a = 1.0
    X = []
    X.append([-0.5*a, 0, 0])
    X.append([0, -0.5*np.sqrt(3)*a, 0])
    X.append([0.5*a, 0, 0])
    X.append([0, 0.5*np.sqrt(3)*a, 0])
    p = geom.add_polygon(X=X, lcar = h)
    geom.add_physical(p.surface, label=100)
    for i in range(4): geom.add_physical(p.line_loop.lines[i], label=1000 + i)
    return geom



import unittest
import numpy as np
import simfempy.applications.heat as heat

#================================================================#
class TestAnalytical(unittest.TestCase):
    def _check(self, results):
        for meth,err in results.items():
            if isinstance(err, dict):
                for m, e in err.items():
                    if not np.all(e<1e-10): raise ValueError("error in method '{}' '{}' error is {}".format(meth,m,e))
            else:
                if not np.all(err<1e-10): raise ValueError("error in method '{}' error is {}".format(meth,err))
#---------------------------
    def test_poisson1d(self):
        from heat.poisson_analytic import test_analytic
        self._check(test_analytic(exactsolution = 'Linear', geomname = "unitline", verbose=0))
    def test_poisson2d(self):
        from heat.poisson_analytic import test_analytic
        self._check(test_analytic(exactsolution = 'Linear', geomname = "unitsquare", verbose=0))
    def test_poisson3d(self):
        from heat.poisson_analytic import test_analytic
        self._check(test_analytic(exactsolution = 'Linear', geomname = "unitcube", verbose=0))
#     # ---------------------------
#     def test_elliptic2d(self):
#         from elliptic.elliptic import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitsquare", verbose=0))
#     def test_elliptic3d(self):
#         from elliptic.elliptic import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitcube", verbose=0))
#     # ---------------------------
#     def test_elasticity2d(self):
#         from elasticity.analytic import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitsquare", verbose=0))
#     def test_elasticity3d(self):
#         from elasticity.analytic import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitcube", verbose=0))
#     # ---------------------------
#     def test_mixedlaplace2d(self):
#         from mixed.laplace import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitsquare", verbose=0))
#     def test_mixedlaplace3d(self):
#         from mixed.laplace import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitcube", verbose=0))
#     # ---------------------------
#     def test_stokes2d(self):
#         from flow.stokes import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitsquare", verbose=0))
#     def test_stokes3d(self):
#         from flow.stokes import test_analytic
#         self._check(test_analytic(exactsolution = 'Linear', geomname = "unitcube", verbose=0))


#================================================================#
if __name__ == '__main__':
    unittest.main(verbosity=2)


import numpy as np
import meshio
import subprocess
from . import simplexmesh

#----------------------------------------------------------------#
def gmshRefine(mesh):
    filenamemsh = "coarse.msh"
    filenamemshref = "fine.msh"
    mesh.write(filename=filenamemsh)
    cmd = ['gmsh']
    cmd += ["-refine"]
    cmd += ["{}".format(filenamemsh)]
    cmd += ["-o"]
    cmd += ["{}".format(filenamemshref)]
    # cmd += ["-format"]
    # cmd += ["{}".format("msh2")]
    p = subprocess.Popen(cmd, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    # if stderr != "":
    #     print("cmd=", cmd)
    #     print("cmd=", ' '.join(cmd))
    #     print("stderr=", stderr)
    #     raise RuntimeError('Gmsh exited with error (return code %d).' % p.returncode)
    mesh = meshio.read(filenamemshref)
    print("mesh.cells", mesh.cells.keys())
    print("mesh.cell_data", mesh.cell_data.keys())
    data =  mesh.points, mesh.cells, mesh.point_data, mesh.cell_data, mesh.field_data
    return simplexmesh.SimplexMesh(mesh=mesh)


#----------------------------------------------------------------#
def add_circle(
        geom,
        x0,
        radius,
        spacing=None,
        lcars=None,
        h=None,
        R=None,
        compound=False,
        num_sections=3,
        holes=None,
        make_surface=True,
):
    """Add circle in the :math:`x`-:math:`y`-plane.
    """
    if holes is None:
        holes = []
    else:
        assert make_surface

    # Define points that make the circle (midpoint and the four cardinal
    # directions).
    if spacing is None:
        spacing = np.arange(num_sections)/ num_sections
    # print("spacing",spacing)
    X = np.zeros((num_sections + 1, len(x0)))
    if num_sections == 4:
        # For accuracy, the points are provided explicitly.
        X[1:, [0, 1]] = np.array(
            [[radius, 0.0], [0.0, radius], [-radius, 0.0], [0.0, -radius]]
        )
    else:
        X[1:, [0, 1]] = np.array(
            [
                [
                    # radius * np.cos(2 * np.pi * k / num_sections),
                    # radius * np.sin(2 * np.pi * k / num_sections),
                    radius * np.cos(2 * np.pi * spacing[k]),
                    radius * np.sin(2 * np.pi * spacing[k]),
                ]
                for k in range(num_sections)
            ]
        )

    if R is not None:
        assert np.allclose(
            abs(np.linalg.eigvals(R)), np.ones(X.shape[1])
        ), "The transformation matrix doesn't preserve circles; at least one eigenvalue lies off the unit circle."
        X = np.dot(X, R.T)

    X += x0

    # Add Gmsh Points.
    p = [geom.add_point(x, lcar=lcars[i]) for i,x in enumerate(X)]

    # Define the circle arcs.
    arcs = [geom.add_circle_arc(p[k], p[0], p[k + 1]) for k in range(1, len(p) - 1)]
    arcs.append(geom.add_circle_arc(p[-1], p[0], p[1]))

    if compound:
        arcs = [geom.add_compound_line(arcs)]

    line_loop = geom.add_line_loop(arcs)

    if make_surface:
        plane_surface = geom.add_plane_surface(line_loop, holes)
    else:
        plane_surface = None

    class Circle(object):
        def __init__(
                self,
                x0,
                radius,
                R,
                compound,
                num_sections,
                holes,
                line_loop,
                plane_surface,
                lcar=None,
        ):
            self.x0 = x0
            self.radius = radius
            self.lcar = lcar
            self.R = R
            self.compound = compound
            self.num_sections = num_sections
            self.holes = holes
            self.line_loop = line_loop
            self.plane_surface = plane_surface
            return

    return Circle(
        x0,
        radius,
        R,
        compound,
        num_sections,
        holes,
        line_loop,
        plane_surface,
    )


#----------------------------------------------------------------#
# def add_holes(geom, x0, x1, **kwargs):
#     h = kwargs.pop('h')
#     hhole = kwargs.pop('hhole')
#     nholes = kwargs.pop('nholes')
#     holesize = kwargs.pop('holesize')
#     if 'hole_labels' in kwargs: hole_labels = kwargs.pop('hole_labels')
#     else: hole_labels = None
#     if 'make_surface' in kwargs: make_surface = kwargs.pop('make_surface')
#     else: make_surface = True
#     spacesize = (x1-x0-nholes*holesize)/(nholes+1)
#     if spacesize < 0.1*holesize:
#         maxsize = (x1-x0)/(nholes*1.1 - 0.1)
#         raise ValueError("holes too big (max={})".format(maxsize))
#     pos = np.empty(2*nholes)
#     pos[0] = spacesize
#     pos[1] = pos[0] + holesize
#     for i in range(1,nholes):
#         pos[2*i] = pos[2*i-1] + spacesize
#         pos[2*i+1] = pos[2*i] + holesize
#     xholes = []
#     for i in range(nholes):
#         xa, xb = x0+pos[2*i], x0+pos[2*i+1]
#         for j in range(nholes):
#             ya, yb = x0+pos[2*j], x0+pos[2*j+1]
#             xholes.append([[xa, ya, 0], [xb, ya, 0], [xb, yb, 0], [xa, yb, 0]])
#
#     holes = []
#     hole_labels = np.arange(200, 200 + len(xholes), dtype=int)
#     for xhole, hole_label in zip(xholes, hole_labels):
#         holes.append(geom.add_polygon(X=xhole, lcar=hhole))
#         xarrm = np.mean(np.array(xhole), axis=0)
#         add_point_in_surface(geom, holes[-1].surface, xarrm, lcar=h)
#         geom.add_physical(holes[-1].surface, label=int(hole_label))
#     return holes, hole_labels
#----------------------------------------------------------------#
def add_holesnew(geom, **kwargs):
    h = kwargs.pop('h')
    hhole = kwargs.pop('hhole')
    x0 = kwargs.pop('x0')
    x1 = kwargs.pop('x1')
    y0 = kwargs.pop('y0')
    y1 = kwargs.pop('y1')
    nholesx = kwargs.pop('nholesx')
    nholesy = kwargs.pop('nholesy')
    if 'holesizex' in kwargs: holesizex = kwargs.pop('holesizex')
    else: holesizex=None
    if 'holesizey' in kwargs: holesizey = kwargs.pop('holesizey')
    else: holesizey=None
    if 'hole_labels' in kwargs: hole_labels = kwargs.pop('hole_labels')
    else: hole_labels = None
    if 'make_surface' in kwargs: make_surface = kwargs.pop('make_surface')
    else: make_surface = True
    holesizexmax = (x1 - x0) / (nholesx * 1.1 - 0.1)
    if holesizex is None: holesizex = holesizexmax*0.99
    if nholesx>1:
        spacesizex = (x1-x0-nholesx*holesizex)/(nholesx-1)
        if spacesizex < 0.1*holesizex:
            raise ValueError("holesizex({}) too big (max={})".format(holesizex,holesizexmax))
    holesizeymax = (y1 - y0) / (nholesy * 1.1 - 0.1)
    if holesizey is None: holesizey = holesizeymax*0.99
    if nholesy>1:
        spacesizey = (y1-y0-nholesy*holesizey)/(nholesy-1)
        if spacesizey < 0.1*holesizey:
            raise ValueError("holesizey({}) too big (max={})".format(holesizey,holesizeymax))
    posx = np.empty(2*nholesx)
    posx[0] = x0
    posx[1] = posx[0] + holesizex
    for i in range(1,nholesx):
        posx[2*i] = posx[2*i-1] + spacesizex
        posx[2*i+1] = posx[2*i] + holesizex
    posy = np.empty(2*nholesy)
    posy[0] = y0
    posy[1] = posy[0] + holesizey
    for i in range(1,nholesy):
        posy[2*i] = posy[2*i-1] + spacesizey
        posy[2*i+1] = posy[2*i] + holesizey

    xholes = []
    for i in range(nholesx):
        xa, xb = posx[2*i], posx[2*i+1]
        for j in range(nholesy):
            ya, yb = posy[2*j], posy[2*j+1]
            xholes.append([[xa, ya, 0], [xb, ya, 0], [xb, yb, 0], [xa, yb, 0]])

    holes = []
    hole_labels = np.arange(200, 200 + len(xholes), dtype=int)
    for xhole, hole_label in zip(xholes, hole_labels):
        holes.append(geom.add_polygon(X=xhole, lcar=hhole))
        xarrm = np.mean(np.array(xhole), axis=0)
        add_point_in_surface(geom, holes[-1].surface, xarrm, lcar=h)
        geom.add_physical(holes[-1].surface, label=int(hole_label))
    return holes, hole_labels

#----------------------------------------------------------------#
def add_polygon(geom, X, lcar=None, holes=None, make_surface=True):
    assert len(X) == len(lcar)
    if holes is None:
        holes = []
    else:
        assert make_surface

    # Create points.
    p = [geom.add_point(x, lcar=l) for x,l in zip(X,lcar)]
    # Create lines
    lines = [geom.add_line(p[k], p[k + 1]) for k in range(len(p) - 1)]
    lines.append(geom.add_line(p[-1], p[0]))
    ll = geom.add_line_loop((lines))
    surface = geom.add_plane_surface(ll, holes) if make_surface else None

    print("ll", ll)
    print("surface", surface)
    return geom.Polygon(ll, surface)
    return geom.Polygon(ll, surface, lcar=lcar)

#----------------------------------------------------------------#
def add_point_in_surface(geom, surf, X, lcar, label=None):
    p = geom.add_point(X, lcar=lcar)
    geom.add_raw_code("Point {{{}}} In Surface {{{}}};".format(p.id, surf.id))
    if label:
        geom.add_physical_point(p, label=label)



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np
import scipy.linalg as linalg
import scipy.sparse as sparse

from simfempy.fems import femp1

try:
    from simfempy.meshes.simplexmesh import SimplexMesh
except ModuleNotFoundError:
    from ..meshes.simplexmesh import SimplexMesh
import simfempy.fems.bdrydata

# =================================================================#
class FemP1(femp1.FemP1):
    def __init__(self, mesh=None):
        super().__init__(mesh)

    def setMesh(self, mesh, ncomp):
        super().setMesh(mesh)
        self.ncomp = ncomp
        ncells, simps = self.mesh.ncells, self.mesh.simplices
        nlocncomp = ncomp * self.nloc
        self.rowssys = np.repeat(ncomp * simps, ncomp).reshape(ncells * self.nloc, ncomp) + np.arange(ncomp)
        self.rowssys = self.rowssys.reshape(ncells, nlocncomp).repeat(nlocncomp).reshape(ncells, nlocncomp, nlocncomp)
        self.colssys = self.rowssys.swapaxes(1, 2)
        self.colssys = self.colssys.reshape(-1)
        self.rowssys = self.rowssys.reshape(-1)

    def prepareBoundary(self, bdrycond, postproc):
        if not isinstance(bdrycond, (list, tuple)):
            return super().prepareBoundary(bdrycond.colorsOfType("Dirichlet"), postproc)
        bdrydata = []
        for icomp in range(self.ncomp):
            bdrydata.append(super().prepareBoundary(bdrycond[icomp].colorsOfType("Dirichlet"), postproc[icomp]))
        return bdrydata

    def computeRhs(self, u, rhs, diff, bdrycond, method, bdrydata):
        b = np.zeros(self.mesh.nnodes * self.ncomp)
        if rhs:
            for icomp in range(self.ncomp):
                x, y, z = self.mesh.points.T
                bnodes = rhs[icomp](x, y, z, diff[icomp][0])
                b[icomp::self.ncomp] = self.massmatrix * bnodes
        normals = self.mesh.normals
        scale = 1 / self.mesh.dimension
        for color, faces in self.mesh.bdrylabels.items():
            for icomp in range(self.ncomp):
                if bdrycond[icomp].type[color] != "Neumann": continue
                normalsS = normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                kS = diff[icomp][self.mesh.cellsOfFaces[faces, 0]]
                x1, y1, z1 = self.mesh.pointsf[faces].T
                nx, ny, nz = normalsS[:, 0] / dS, normalsS[:, 1] / dS, normalsS[:, 2] / dS
                bS = scale * dS * bdrycond[icomp].fct[color](x1, y1, z1, nx, ny, nz, kS)
                indices = icomp + self.ncomp * self.mesh.faces[faces]
                np.add.at(b, indices.T, bS)
        return self.vectorDirichlet(b, u, bdrycond, method, bdrydata)

    def matrixDiffusion(self, k, bdrycond, method, bdrydata):
        nnodes, ncells, ncomp = self.mesh.nnodes, self.mesh.ncells, self.ncomp
        matxx = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 0], self.cellgrads[:, :, 0])
        matyy = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 1], self.cellgrads[:, :, 1])
        matzz = np.einsum('nk,nl->nkl', self.cellgrads[:, :, 2], self.cellgrads[:, :, 2])
        nlocncomp = ncomp * self.nloc
        mat = np.zeros(shape=(ncells, nlocncomp, nlocncomp))
        for icomp in range(ncomp):
            mat[:, icomp::ncomp, icomp::ncomp] = ((matxx + matyy + matzz).T * self.mesh.dV * k[icomp]).T
        A = sparse.coo_matrix((mat.ravel(), (self.rowssys, self.colssys)), shape=(ncomp*nnodes, ncomp*nnodes)).tocsr()
        return self.matrixDirichlet(A, bdrycond, method, bdrydata)

    def vectorDirichlet(self, b, u, bdrycond, method, bdrydata):
        x, y, z = self.mesh.points.T
        if u is None: u = np.zeros_like(b)
        else: assert u.shape == b.shape
        nnodes, ncomp = self.mesh.nnodes, self.ncomp
        for icomp in range(ncomp):
            nodesdir, nodedirall, nodesinner, nodesdirflux = bdrydata[icomp].nodesdir, bdrydata[icomp].nodedirall, bdrydata[icomp].nodesinner, bdrydata[icomp].nodesdirflux
            for key, nodes in nodesdirflux.items():
                bdrydata[icomp].bsaved[key] = b[icomp + ncomp * nodes]
            if method == 'trad':
                for color, nodes in nodesdir.items():
                    dirichlet = bdrycond[icomp].fct[color]
                    b[icomp + ncomp * nodes] = dirichlet(x[nodes], y[nodes], z[nodes])
                    u[icomp + ncomp * nodes] = b[icomp + ncomp * nodes]
                indin = icomp + ncomp *nodesinner
                inddir = icomp + ncomp *nodedirall
                b[indin] -= bdrydata[icomp].A_inner_dir * b[inddir]
            else:
                for color, nodes in nodesdir.items():
                    dirichlet = bdrycond[icomp].fct[color]
                    u[icomp + ncomp * nodes] = dirichlet(x[nodes], y[nodes], z[nodes])
                    b[icomp + ncomp * nodes] = 0
                indin = icomp + ncomp *nodesinner
                inddir = icomp + ncomp *nodedirall
                b[indin] -= bdrydata[icomp].A_inner_dir * u[inddir]
                b[inddir] = bdrydata[icomp].A_dir_dir * u[inddir]
        return b, u, bdrydata

    def matrixDirichlet(self, A, bdrycond, method, bdrydata):
        nnodes, ncomp = self.mesh.nnodes, self.ncomp
        for icomp in range(ncomp):
            nodesdir, nodedirall, nodesinner, nodesdirflux = bdrydata[icomp].nodesdir, bdrydata[icomp].nodedirall, bdrydata[icomp].nodesinner, bdrydata[icomp].nodesdirflux
            indin = icomp + ncomp * nodesinner
            inddir = icomp + ncomp * nodedirall
            bdrydata[icomp].A_inner_dir = A[indin, :][:, inddir]
            for key, nodes in nodesdirflux.items():
                nb = nodes.shape[0]
                help = sparse.dok_matrix((nb, ncomp * nnodes))
                for i in range(nb): help[i, icomp + ncomp * nodes[i]] = 1
                bdrydata[icomp].Asaved[key] = help.dot(A)
            if method == 'trad':
                help = np.ones((ncomp * nnodes))
                help[inddir] = 0
                help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                A = help.dot(A.dot(help))
                help = np.zeros((ncomp * nnodes))
                help[inddir] = 1.0
                help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                A += help
            else:
                bdrydata[icomp].A_dir_dir = A[inddir, :][:, inddir]
                help = np.ones((ncomp * nnodes))
                help[inddir] = 0
                help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                help2 = np.zeros((ncomp * nnodes))
                help2[inddir] = 1
                help2 = sparse.dia_matrix((help2, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
        return A, bdrydata

    def boundary(self, A, b, u, bdrycond, bdrydata, method):
        x, y, z = self.mesh.points.T
        nnodes, ncomp = self.mesh.nnodes, self.ncomp
        self.bsaved = []
        self.Asaved = []
        for icomp in range(ncomp):
            nodedirall, nodesinner, nodesdir, nodesdirflux = bdrydata[icomp]
            self.bsaved.append({})
            self.Asaved.append({})
            for key, nodes in nodesdirflux.items():
                self.bsaved[icomp][key] = b[icomp + ncomp * nodes]
            for key, nodes in nodesdirflux.items():
                nb = nodes.shape[0]
                help = sparse.dok_matrix((nb, ncomp * nnodes))
                for i in range(nb): help[i, icomp + ncomp * nodes[i]] = 1
                self.Asaved[icomp][key] = help.dot(A)
            if method == 'trad':
                for color, nodes in nodesdir.items():
                    dirichlet = bdrycond[icomp].fct[color]
                    b[icomp + ncomp * nodes] = dirichlet(x[nodes], y[nodes], z[nodes])
                    u[icomp + ncomp * nodes] = b[icomp + ncomp * nodes]
                indin = icomp + ncomp *nodesinner
                inddir = icomp + ncomp *nodedirall
                b[indin] -= A[indin, :][:,inddir] * b[inddir]
                help = np.ones((ncomp * nnodes))
                help[inddir] = 0
                help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                A = help.dot(A.dot(help))
                help = np.zeros((ncomp * nnodes))
                help[inddir] = 1.0
                help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                A += help
            else:
                for color, nodes in nodesdir.items():
                    dirichlet = bdrycond[icomp].fct[color]
                    u[icomp + ncomp * nodes] = dirichlet(x[nodes], y[nodes], z[nodes])
                    b[icomp + ncomp * nodes] = 0
                # print("b", b)
                # print("u", u)
                # b -= A*u
                indin = icomp + ncomp *nodesinner
                inddir = icomp + ncomp *nodedirall
                b[indin] -= A[indin, :][:, inddir] * u[inddir]
                b[inddir] = A[inddir, :][:, inddir] * u[inddir]
                # print("b", b)
                help = np.ones((ncomp * nnodes))
                help[inddir] = 0
                # print("help", help)
                help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                help2 = np.zeros((ncomp * nnodes))
                help2[inddir] = 1
                # print("help2", help2)
                help2 = sparse.dia_matrix((help2, 0), shape=(ncomp * nnodes, ncomp * nnodes))
                A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
                # print("A", A)
        return A, b, u

    def tonode(self, u):
        return u

    def grad(self, ic):
        normals = self.mesh.normals[self.mesh.facesOfCells[ic, :]]
        grads = 0.5 * normals / self.mesh.dV[ic]
        chsg = (ic == self.mesh.cellsOfFaces[self.mesh.facesOfCells[ic, :], 0])
        # print("### chsg", chsg, "normals", normals)
        grads[chsg] *= -1.
        return grads

    def phi(self, ic, x, y, z, grad):
        return 1. / 3. + np.dot(grad, np.array(
            [x - self.mesh.pointsc[ic, 0], y - self.mesh.pointsc[ic, 1], z - self.mesh.pointsc[ic, 2]]))

    def testgrad(self):
        for ic in range(fem.mesh.ncells):
            grads = fem.grad(ic)
            for ii in range(3):
                x = self.mesh.points[self.mesh.simplices[ic, ii], 0]
                y = self.mesh.points[self.mesh.simplices[ic, ii], 1]
                z = self.mesh.points[self.mesh.simplices[ic, ii], 2]
                for jj in range(3):
                    phi = self.phi(ic, x, y, z, grads[jj])
                    if ii == jj:
                        test = np.abs(phi - 1.0)
                        if test > 1e-14:
                            print('ic=', ic, 'grad=', grads)
                            print('x,y', x, y)
                            print('x-xc,y-yc', x - self.mesh.pointsc[ic, 0], y - self.mesh.pointsc[ic, 1])
                            raise ValueError('wrong in cell={}, ii,jj={},{} test= {}'.format(ic, ii, jj, test))
                    else:
                        test = np.abs(phi)
                        if np.abs(phi) > 1e-14:
                            print('ic=', ic, 'grad=', grads)
                            raise ValueError('wrong in cell={}, ii,jj={},{} test= {}'.format(ic, ii, jj, test))

    def computeErrorL2(self, solex, uh):
        x, y, z = self.mesh.points.T
        e = []
        err = []
        for icomp in range(self.ncomp):
            e.append(solex[icomp](x, y, z) - uh[icomp::self.ncomp])
            err.append(np.sqrt(np.dot(e[icomp], self.massmatrix * e[icomp])))
        return err, e

    def computeBdryMean(self, u, data, icomp=None):
        colors = [int(x) for x in data.split(',')]
        if icomp is None:
            mean, omega = np.zeros(shape=(self.ncomp,len(colors))), np.zeros(len(colors))
            for i,color in enumerate(colors):
                faces = self.mesh.bdrylabels[color]
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                omega[i] = np.sum(dS)
                for icomp in range(self.ncomp):
                    mean[icomp, i] = np.sum(dS * np.mean(u[icomp + self.ncomp * self.mesh.faces[faces]], axis=1))
            return mean / omega
        else:
            mean, omega = np.zeros(shape=(len(colors))), np.zeros(len(colors))
            for i,color in enumerate(colors):
                faces = self.mesh.bdrylabels[color]
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                omega[i] = np.sum(dS)
                mean[i] = np.sum(dS * np.mean(u[icomp + self.ncomp * self.mesh.faces[faces]], axis=1))
            return mean/omega

    def computeBdryDn(self, u, data, bdrydata, bdrycond, icomp=None):
        colors = [int(x) for x in data.split(',')]
        if icomp is None:
            flux, omega = np.zeros(shape=(self.ncomp,len(colors))), np.zeros(len(colors))
            for i,color in enumerate(colors):
                faces = self.mesh.bdrylabels[color]
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                omega[i] = np.sum(dS)
                for icomp in range(self.ncomp):
                    if bdrycond[icomp].type[color] == "Dirichlet":
                        bs, As = bdrydata[icomp].bsaved[color], bdrydata[icomp].Asaved[color]
                        flux[icomp, i] = np.sum(As * u - bs)
                else:
                    raise NotImplementedError("computeBdryDn for condition '{}'".format(bdrycond.type[color]))
            return flux
        else:
            flux, omega = np.zeros(shape=(len(colors))), np.zeros(len(colors))
            for i,color in enumerate(colors):
                faces = self.mesh.bdrylabels[color]
                normalsS = self.mesh.normals[faces]
                dS = linalg.norm(normalsS, axis=1)
                omega[i] = np.sum(dS)
                if bdrycond[icomp].type[color] == "Dirichlet":
                    bs, As = bdrydata[icomp].bsaved[color], bdrydata[icomp].Asaved[color]
                    flux[i] = np.sum(As * u - bs)
                else:
                    raise NotImplementedError("computeBdryDn for condition '{}'".format(bdrycond.type[color]))
            return flux


    def computePointValues(self, u, data):
        colors = [int(x) for x in data.split(',')]
        up = np.empty(shape=(len(colors), self.ncomp))
        for i,color in enumerate(colors):
            nodes = self.mesh.verticesoflabel[color]
            for icomp in range(self.ncomp):
                up[i, icomp] = u[icomp + self.ncomp *nodes]
        return up

    def computeMeanValues(self, u, data):
        colors = [int(x) for x in data.split(',')]
        up = np.empty(shape=(len(colors), self.ncomp))
        for i, color in enumerate(colors):
            cells = self.mesh.cellsoflabel[color]
            simpcells = self.mesh.simplices[cells]
            for icomp in range(self.ncomp):
                mean = np.sum(np.mean(u[icomp + self.ncomp * simpcells], axis=1) * self.mesh.dV[cells])
                up[i, icomp] = mean
        return up

# ------------------------------------- #

if __name__ == '__main__':
    trimesh = SimplexMesh(geomname="backwardfacingstep", hmean=0.3)
    fem = FemP1(trimesh)
    fem.testgrad()
    import plotmesh
    import matplotlib.pyplot as plt

    plotmesh.meshWithBoundaries(trimesh)
    plt.show()



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np
import scipy.linalg as linalg
import scipy.sparse as sparse
try:
    from simfempy.meshes.simplexmesh import SimplexMesh
except ModuleNotFoundError:
    from ..meshes.simplexmesh import SimplexMesh
import simfempy.fems.bdrydata

#=================================================================#
class FemBV0(object):
    """
    on suppose que  self.mesh.edgesOfCell[ic, kk] et oppose à elem[ic,kk] !!!
    """
    def __init__(self, mesh=None, massproj=False):
        if mesh is not None:
            self.setMesh(mesh)
        self.massproj=massproj

    def setMesh(self, mesh):
        self.mesh = mesh
        self.nloc = self.mesh.dimension+1
        self.Mtocell = self.toCellMatrix()

    def toCellMatrix(self):
        ncells, nfaces, normals, sigma, facesofcells = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals, self.mesh.sigma, self.mesh.facesOfCells
        dim, dV, nloc, p, pc, simp = self.mesh.dimension, self.mesh.dV, self.nloc, self.mesh.points, self.mesh.pointsc, self.mesh.simplices

        # pour RT
        dS = sigma * linalg.norm(normals[facesofcells], axis=2)/dim
        ps = p[simp][:,:,:dim]
        ps2 = np.transpose(ps, axes=(2,0,1))
        pc2 = np.repeat(pc[:,:dim].T[:, :, np.newaxis], nloc, axis=2)
        pd = pc2 -ps2
        rows = np.repeat((np.repeat(dim * np.arange(ncells), dim).reshape(ncells,dim) + np.arange(dim)).swapaxes(1,0),nloc)
        cols = np.tile(facesofcells.ravel(), dim)
        mat = np.einsum('ni, jni, n->jni', dS, pd, 1/dV)
        return  sparse.coo_matrix((mat.ravel(), (rows.ravel(), cols.ravel())), shape=(dim*ncells, nfaces))


        # pour les tilde
        # rows = np.repeat((np.repeat(dim * np.arange(ncells), dim).reshape(ncells,dim) + np.arange(dim)).swapaxes(1,0),nloc)
        # cols = np.tile(facesofcells.ravel(), dim)
        # dS = linalg.norm(normals[facesofcells], axis=2)
        # mat = np.einsum('nij, ni, n->jni', normals[facesofcells][:,:,:dim], 1/dS, dV/(dim+1))
        # return  sparse.coo_matrix((mat.ravel(), (rows.ravel(), cols.ravel())), shape=(dim*ncells, nfaces))

        # pour les hat
        dS = sigma * linalg.norm(normals[facesofcells], axis=2)
        ps = p[simp][:,:,:dim]
        ps2 = np.transpose(ps, axes=(2,0,1))
        pc2 = np.repeat(pc[:,:dim].T[:, :, np.newaxis], nloc, axis=2)
        pd = pc2 -ps2
        rows = np.repeat((np.repeat(dim * np.arange(ncells), dim).reshape(ncells,dim) + np.arange(dim)).swapaxes(1,0),nloc)
        cols = np.tile(facesofcells.ravel(), dim)
        scale = 1/dim
        mat = scale*np.einsum('ni, jni->jni', dS, pd)
        return  sparse.coo_matrix((mat.ravel(), (rows.ravel(), cols.ravel())), shape=(dim*ncells, nfaces))

    def toCell(self, v):
        return self.Mtocell.dot(v)

    def constructMass(self, diffinvcell=None):
        # ncells, nfaces, normals = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals
        # cellsOfFaces, facesOfCells, dV = self.mesh.cellsOfFaces, self.mesh.facesOfCells, self.mesh.dV
        # dim, nloc, p, pc, simp = self.mesh.dimension, self.nloc, self.mesh.points, self.mesh.pointsc, self.mesh.simplices
        #
        # assert dim == 2
        # dS2 = np.einsum('nik,nik->ni', normals[facesOfCells], normals[facesOfCells])
        # ps = p[simp][:, :, :dim]
        # ps2 = np.transpose(ps, axes=(2, 0, 1))
        # pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
        # pd = pc2 - ps2
        # # print("pd.shape", pd.shape)
        # a = (dim + 1) / dim ** 3
        # mat = a * dS2 * np.einsum('kni,kni,n->ni', pd, pd, diffinvcell / dV)
        # # print("pd**2", np.einsum('kni,kni->ni', pd, pd))
        # # print("dS2", dS2)
        # # print("mat", mat)
        # rows = facesOfCells.ravel()
        # A = sparse.coo_matrix((mat.ravel(), (rows, rows)), shape=(nfaces, nfaces)).tocsr()
        # print("A", A)
        # # self.mesh.plotWithNumbering()
        # return A

        ncells, nfaces, normals, sigma, facesofcells = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals, self.mesh.sigma, self.mesh.facesOfCells
        dim, dV, nloc, simp = self.mesh.dimension, self.mesh.dV, self.nloc, self.mesh.simplices
        p, pc, pf = self.mesh.points, self.mesh.pointsc, self.mesh.pointsf

        dS = sigma*linalg.norm(normals[facesofcells], axis=2)
        ps = p[simp][:, :, :dim]
        ps2 = np.transpose(ps, axes=(2, 0, 1))
        pc2 = np.repeat(pc[:, :dim].T[:, :, np.newaxis], nloc, axis=2)
        pd = pc2 - ps2
        pf2 = pf[facesofcells][:,:,:dim]
        # print("pd", pd.shape)
        # print("pf2", pf2.shape)
        # print("dS", dS.shape)
        scale = 1/dim/dim

        mat  = np.einsum('kni, nik, nj, ni, n->nij', pd, pf2, dS, dS, 1/dV)
        mat -= np.einsum('kni, njk, nj, ni, n->nij', pd, ps, dS, dS, 1/dV)
        # mat  = np.einsum('kni, knj, ni, nj, n->nij', pd, pd, dS, dS, 1/dV)

        mat *= scale

        rows = np.repeat(facesofcells, self.nloc).ravel()
        cols = np.tile(facesofcells, self.nloc).ravel()
        A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(nfaces, nfaces)).tocsr()
        # print("A (BV)", A)
        return A

        ncells, nfaces, normals, sigma, facesofcells = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals, self.mesh.sigma, self.mesh.facesOfCells
        dim, dV, nloc, p, pc, simp = self.mesh.dimension, self.mesh.dV, self.nloc, self.mesh.points, self.mesh.pointsc, self.mesh.simplices
        pf = self.mesh.pointsf[facesofcells][:,:,:dim]
        ps = p[simp][:, :, :dim]
        pd = pf - ps
        # print("pd", pd)
        # print("normal", normals[facesofcells][:,:,:dim])
        scale = 1/dim/(dim+1)
        mat = scale*np.einsum('nij, nij, ni->ni', pd, normals[facesofcells][:,:,:dim], sigma)
        rows = facesofcells.ravel()
        A = sparse.coo_matrix((mat.ravel(), (rows, rows)), shape=(nfaces, nfaces)).tocsr()
        # print("A", A)
        # stop
        return A



        dS = sigma * linalg.norm(normals[facesofcells], axis=2)
        ps = p[simp][:,:,:dim]
        ps2 = np.transpose(ps, axes=(2,0,1))
        pc2 = np.repeat(pc[:,:dim].T[:, :, np.newaxis], nloc, axis=2)
        pd = pc2 -ps2
        rows = np.repeat((np.repeat(dim * np.arange(ncells), dim).reshape(ncells,dim) + np.arange(dim)).swapaxes(1,0),nloc)
        cols = np.tile(facesofcells.ravel(), dim)
        # scale = (dim+1)/dim**3 * 4/5
        scale = (dim+1)/dim**2
        mat = scale*np.einsum('ni, ni, jni, jni, n->ni', dS, dS, pd, pd, diffinvcell/dV)
        rows = facesofcells.ravel()
        A = sparse.coo_matrix((mat.ravel(), (rows, rows)), shape=(nfaces, nfaces)).tocsr()
        # print("A", A)
        return A


    def constructDiv(self):
        ncells, nfaces, normals, sigma, facesofcells = self.mesh.ncells, self.mesh.nfaces, self.mesh.normals, self.mesh.sigma, self.mesh.facesOfCells
        rows = np.repeat(np.arange(ncells), self.nloc)
        cols = facesofcells.ravel()
        mat =  (sigma*linalg.norm(normals[facesofcells],axis=2)).ravel()
        return  sparse.coo_matrix((mat, (rows, cols)), shape=(ncells, nfaces)).tocsr()

    def reconstruct(self, p, vc, diffinv):
        nnodes, ncells, dim = self.mesh.nnodes, self.mesh.ncells, self.mesh.dimension
        if len(diffinv.shape) != 1:
            raise NotImplemented("only scalar diffusion the time being")
        counts = np.bincount(self.mesh.simplices.reshape(-1))
        pn2 = np.zeros(nnodes)
        xdiff = self.mesh.points[self.mesh.simplices, :dim] - self.mesh.pointsc[:, np.newaxis,:dim]
        rows = np.repeat(self.mesh.simplices,dim)
        cols = np.repeat(dim*np.arange(ncells),dim*(dim+1)).reshape(ncells * (dim+1), dim) + np.arange(dim)
        mat = np.einsum("nij, n -> nij", xdiff, diffinv)
        A = sparse.coo_matrix((mat.reshape(-1), (rows.reshape(-1), cols.reshape(-1))), shape=(nnodes, dim*ncells)).tocsr()
        np.add.at(pn2, self.mesh.simplices.T, p)
        pn2 += A*vc
        pn2 /= counts
        return pn2


    def rhsDirichlet(self, faces, ud):
        scale = 14/15
        scale = 1
        return linalg.norm(self.mesh.normals[faces],axis=1) * ud * scale

    def constructRobin(self, bdrycond, type):
        nfaces = self.mesh.nfaces
        rows = np.empty(shape=(0), dtype=int)
        cols = np.empty(shape=(0), dtype=int)
        mat = np.empty(shape=(0), dtype=float)
        for color, faces in self.mesh.bdrylabels.items():
            if bdrycond.type[color] != type: continue
            if not bdrycond.param[color]: continue
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            cols = np.append(cols, faces)
            rows = np.append(rows, faces)
            mat = np.append(mat, 1/bdrycond.param[color] * dS)
        return sparse.coo_matrix((mat, (rows, cols)), shape=(nfaces, nfaces)).tocsr()

    def matrixNeumann(self, A, B, bdrycond):
        nfaces = self.mesh.nfaces
        bdrydata = simfempy.fems.bdrydata.BdryData()
        bdrydata.facesneumann = np.empty(shape=(0), dtype=int)
        bdrydata.colorsneum = bdrycond.colorsOfType("Neumann")
        for color in bdrydata.colorsneum:
            bdrydata.facesneumann = np.unique(np.union1d(bdrydata.facesneumann, self.mesh.bdrylabels[color]))
        bdrydata.facesinner = np.setdiff1d(np.arange(self.mesh.nfaces, dtype=int), bdrydata.facesneumann)

        bdrydata.B_inner_neum = B[:, :][:, bdrydata.facesneumann]
        help = np.ones(nfaces)
        help[bdrydata.facesneumann] = 0
        help = sparse.dia_matrix((help, 0), shape=(nfaces, nfaces))
        B = B.dot(help)

        bdrydata.A_inner_neum = A[bdrydata.facesinner, :][:, bdrydata.facesneumann]
        bdrydata.A_neum_neum = A[bdrydata.facesneumann, :][:, bdrydata.facesneumann]
        help2 = np.zeros((nfaces))
        help2[bdrydata.facesneumann] = 1
        help2 = sparse.dia_matrix((help2, 0), shape=(nfaces, nfaces))
        A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))

        return bdrydata, A, B



class BdryData(object):
    """
    Information for boundary conditions
    """
    def __init__(self):
        self.bsaved = {}
        self.Asaved = {}

    def __repr__(self):
        return ", ".join("'{}': {}".format(attr, value) for attr, value in self.__dict__.items())


import numpy as np
import scipy.sparse as sparse
import scipy.linalg as linalg
import scipy.sparse.linalg as splinalg
from simfempy import solvers
from simfempy import fems
from simfempy import tools

#=================================================================#
class Elasticity(solvers.solver.Application):
    """
    """
    YoungPoisson = {}
    YoungPoisson["Acier"] = (210, 0.285)
    YoungPoisson["Aluminium"] = (71, 0.34)
    YoungPoisson["Verre"] = (60, 0.25)
    YoungPoisson["Beton"] = (10, 0.15)
    YoungPoisson["Caoutchouc"] = (0.2, 0.49)
    YoungPoisson["Bois"] = (7, 0.2)
    YoungPoisson["Marbre"] = (26, 0.3)

    def toLame(self, E, nu):
        return 0.5*E/(1+nu), nu*E/(1+nu)/(1-2*nu)

    def material2Lame(self, material):
        E, nu = self.YoungPoisson[material]
        return self.toLame(E, nu)

    def generatePoblemData(self, **kwargs):
        self.ncomp = self.mesh.dimension
        return super().generatePoblemData(**kwargs)

    def defineRhsAnalyticalSolution(self, solexact):
        def _fctu(x, y, z):
            rhs = np.zeros(shape=(self.ncomp, x.shape[0]))
            mu, lam = self.mufct(0), self.lamfct(0)
            for i in range(self.ncomp):
                for j in range(self.ncomp):
                    rhs[i] -= (lam+mu) * solexact[j].dd(i, j, x, y, z)
                    rhs[i] -= mu * solexact[i].dd(j, j, x, y, z)
            return rhs
        return _fctu

    def defineNeumannAnalyticalSolution(self, solexact):
        def _fctneumann(x, y, z, nx, ny, nz):
            rhs = np.zeros(shape=(self.ncomp, x.shape[0]))
            normals = nx, ny, nz
            mu, lam = self.mufct(0), self.lamfct(0)
            for i in range(self.ncomp):
                for j in range(self.ncomp):
                    rhs[i] += lam * solexact[j].d(j, x, y, z) * normals[i]
                    rhs[i] += mu  * solexact[i].d(j, x, y, z) * normals[j]
                    rhs[i] += mu  * solexact[j].d(i, x, y, z) * normals[j]
            return rhs
        return _fctneumann

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # if 'geometry' in kwargs:
        #     return
        self.linearsolver = 'pyamg'
        if 'fem' in kwargs: fem = kwargs.pop('fem')
        else: fem='p1'
        if fem == 'p1':
            self.fem = fems.femp1sys.FemP1()
        elif fem == 'cr1':
            raise NotImplementedError("cr1 not ready")
            self.fem = fems.femcr1sys.FemCR1()
        else:
            raise ValueError("unknown fem '{}'".format(fem))
        if 'material' in kwargs: material = kwargs.pop('material')
        else: material = "Acier"
        self.setParameters(*self.material2Lame(material))
        if 'method' in kwargs: self.method = kwargs.pop('method')
        else: self.method="trad"

    def setParameters(self, mu, lam):
        self.mu, self.lam = mu, lam
        self.mufct = np.vectorize(lambda j: mu)
        self.lamfct = np.vectorize(lambda j: lam)
        if hasattr(self,'mesh'):
            self.mucell = self.mufct(self.mesh.cell_labels)
            self.lamcell = self.lamfct(self.mesh.cell_labels)

    def setMesh(self, mesh):
        super().setMesh(mesh)
        self.fem.setMesh(self.mesh, self.ncomp)
        self.bdrydata = self.fem.prepareBoundary(self.problemdata.bdrycond, self.problemdata.postproc)
        self.mucell = self.mufct(self.mesh.cell_labels)
        self.lamcell = self.lamfct(self.mesh.cell_labels)

    def solve(self, iter, dirname):
        return self.solveLinear()

    def computeRhs(self, u=None):
        ncomp = self.ncomp
        b = np.zeros(self.mesh.nnodes * self.ncomp)
        if self.problemdata.rhs:
            x, y, z = self.mesh.points.T
            rhsall = self.problemdata.rhs(x, y, z)
            for i in range(ncomp):
                b[i::self.ncomp] = self.fem.massmatrix * rhsall[i]
        normals = self.mesh.normals
        for color, faces in self.mesh.bdrylabels.items():
            if self.problemdata.bdrycond.type[color] != "Neumann": continue
            scale = 1 / self.mesh.dimension
            normalsS = normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            xS = np.mean(self.mesh.points[self.mesh.faces[faces]], axis=1)
            x1, y1, z1 = xS[:, 0], xS[:, 1], xS[:, 2]
            nx, ny, nz = normalsS[:, 0] / dS, normalsS[:, 1] / dS, normalsS[:, 2] / dS
            if not color in self.problemdata.bdrycond.fct.keys(): continue
            neumanns = self.problemdata.bdrycond.fct[color](x1, y1, z1, nx, ny, nz)
            for i in range(ncomp):
                bS = scale * dS * neumanns[i]
                indices = i + self.ncomp * self.mesh.faces[faces]
                np.add.at(b, indices.T, bS)
        return self.vectorDirichlet(b, u)


    def matrix(self):
        nnodes, ncells, ncomp, dV = self.mesh.nnodes, self.mesh.ncells, self.ncomp, self.mesh.dV
        nloc, rows, cols, cellgrads = self.fem.nloc, self.fem.rowssys, self.fem.colssys, self.fem.cellgrads
        mat = np.zeros(shape=rows.shape, dtype=float).reshape(ncells, ncomp * nloc, ncomp * nloc)
        for i in range(ncomp):
            for j in range(self.ncomp):
                mat[:, i::ncomp, j::ncomp] += (np.einsum('nk,nl->nkl', cellgrads[:, :, i], cellgrads[:, :, j]).T * dV * self.lamcell).T
                mat[:, i::ncomp, j::ncomp] += (np.einsum('nk,nl->nkl', cellgrads[:, :, j], cellgrads[:, :, i]).T * dV * self.mucell).T
                mat[:, i::ncomp, i::ncomp] += (np.einsum('nk,nl->nkl', cellgrads[:, :, j], cellgrads[:, :, j]).T * dV * self.mucell).T
        A = sparse.coo_matrix((mat.ravel(), (rows, cols)), shape=(ncomp*nnodes, ncomp*nnodes)).tocsr()
        # if self.method == "sym":
        # rows, cols = A.nonzero()
        # A[cols, rows] = A[rows, cols]
        return self.matrixDirichlet(A).tobsr()

    def matrixDirichlet(self, A):
        nnodes, ncomp = self.mesh.nnodes, self.ncomp
        nodesdir, nodedirall, nodesinner, nodesdirflux = self.bdrydata.nodesdir, self.bdrydata.nodedirall, self.bdrydata.nodesinner, self.bdrydata.nodesdirflux
        for key, nodes in nodesdirflux.items():
            nb = nodes.shape[0]
            help = sparse.dok_matrix((ncomp *nb, ncomp * nnodes))
            for icomp in range(ncomp):
                for i in range(nb): help[icomp + ncomp * i, icomp + ncomp * nodes[i]] = 1
            self.bdrydata.Asaved[key] = help.dot(A)
        indin = np.repeat(ncomp * nodesinner, ncomp)
        for icomp in range(ncomp): indin[icomp::ncomp] += icomp
        inddir = np.repeat(ncomp * nodedirall, ncomp)
        for icomp in range(ncomp): inddir[icomp::ncomp] += icomp
        self.bdrydata.A_inner_dir = A[indin, :][:, inddir]
        if self.method == 'trad':
            help = np.ones((ncomp * nnodes))
            help[inddir] = 0
            help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            A = help.dot(A.dot(help))
            help = np.zeros((ncomp * nnodes))
            help[inddir] = 1.0
            help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            A += help
        else:
            self.bdrydata.A_dir_dir = A[inddir, :][:, inddir]
            help = np.ones((ncomp * nnodes))
            help[inddir] = 0
            help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            help2 = np.zeros((ncomp * nnodes))
            help2[inddir] = 1
            help2 = sparse.dia_matrix((help2, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
        return A

    def vectorDirichlet(self, b, u):
        if u is None: u = np.zeros_like(b)
        else: assert u.shape == b.shape
        x, y, z = self.mesh.points.T
        nnodes, ncomp = self.mesh.nnodes, self.ncomp
        nodesdir, nodedirall, nodesinner, nodesdirflux = self.bdrydata.nodesdir, self.bdrydata.nodedirall, self.bdrydata.nodesinner, self.bdrydata.nodesdirflux
        for key, nodes in nodesdirflux.items():
            ind = np.repeat(ncomp * nodes, ncomp)
            for icomp in range(ncomp):ind[icomp::ncomp] += icomp
            self.bdrydata.bsaved[key] = b[ind]
        indin = np.repeat(ncomp * nodesinner, ncomp)
        for icomp in range(ncomp): indin[icomp::ncomp] += icomp
        inddir = np.repeat(ncomp * nodedirall, ncomp)
        for icomp in range(ncomp): inddir[icomp::ncomp] += icomp
        if self.method == 'trad':
            for color, nodes in nodesdir.items():
                if color in self.problemdata.bdrycond.fct.keys():
                    dirichlets = self.problemdata.bdrycond.fct[color](x[nodes], y[nodes], z[nodes])
                    for icomp in range(ncomp):
                        b[icomp + ncomp * nodes] = dirichlets[icomp]
                        u[icomp + ncomp * nodes] = b[icomp + ncomp * nodes]
                else:
                    for icomp in range(ncomp):
                        b[icomp + ncomp * nodes] = 0
                        u[icomp + ncomp * nodes] = b[icomp + ncomp * nodes]
            b[indin] -= self.bdrydata.A_inner_dir * b[inddir]
        else:
            for color, nodes in nodesdir.items():
                if color in self.problemdata.bdrycond.fct.keys():
                    dirichlets = self.problemdata.bdrycond.fct[color](x[nodes], y[nodes], z[nodes])
                    for icomp in range(ncomp):
                        u[icomp + ncomp * nodes] = dirichlets[icomp]
                        b[icomp + ncomp * nodes] = 0
                else:
                    for icomp in range(ncomp):
                        b[icomp + ncomp * nodes] = 0
                        u[icomp + ncomp * nodes] = b[icomp + ncomp * nodes]
            b[indin] -= self.bdrydata.A_inner_dir * u[inddir]
            b[inddir] = self.bdrydata.A_dir_dir * u[inddir]
        return b, u

    def boundary(self, A, b, u=None):
        x, y, z = self.mesh.points.T
        nnodes, ncomp = self.mesh.nnodes, self.ncomp
        nodedirall, nodesinner, nodesdir, nodesdirflux = self.bdrydata
        self.bsaved = {}
        self.Asaved = {}
        for key, nodes in nodesdirflux.items():
            ind = np.repeat(ncomp * nodes, ncomp)
            for icomp in range(ncomp):ind[icomp::ncomp] += icomp
            self.bsaved[key] = b[ind]
        for key, nodes in nodesdirflux.items():
            nb = nodes.shape[0]
            help = sparse.dok_matrix((ncomp *nb, ncomp * nnodes))
            for icomp in range(ncomp):
                for i in range(nb): help[icomp + ncomp * i, icomp + ncomp * nodes[i]] = 1
            self.Asaved[key] = help.dot(A)
        if u is None: u = np.zeros_like(b)
        indin = np.repeat(ncomp * nodesinner, ncomp)
        for icomp in range(ncomp): indin[icomp::ncomp] += icomp
        inddir = np.repeat(ncomp * nodedirall, ncomp)
        for icomp in range(ncomp): inddir[icomp::ncomp] += icomp
        if self.method == 'trad':
            for color, nodes in nodesdir.items():
                if color in self.problemdata.bdrycond.fct.keys():
                    dirichlets = self.problemdata.bdrycond.fct[color](x[nodes], y[nodes], z[nodes])
                    for icomp in range(ncomp):
                        b[icomp + ncomp * nodes] = dirichlets[icomp]
                        u[icomp + ncomp * nodes] = b[icomp + ncomp * nodes]
                else:
                    for icomp in range(ncomp):
                        b[icomp + ncomp * nodes] = 0
                        u[icomp + ncomp * nodes] = b[icomp + ncomp * nodes]
            b[indin] -= A[indin, :][:,inddir] * b[inddir]
            help = np.ones((ncomp * nnodes))
            help[inddir] = 0
            help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            A = help.dot(A.dot(help))
            help = np.zeros((ncomp * nnodes))
            help[inddir] = 1.0
            help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            A += help
        else:
            for color, nodes in nodesdir.items():
                dirichlets = self.problemdata.bdrycond.fct[color](x[nodes], y[nodes], z[nodes])
                for icomp in range(ncomp):
                    u[icomp + ncomp * nodes] = dirichlets[icomp]
                    b[icomp + ncomp * nodes] = 0
            b[indin] -= A[indin, :][:, inddir] * u[inddir]
            b[inddir] = A[inddir, :][:, inddir] * u[inddir]
            help = np.ones((ncomp * nnodes))
            help[inddir] = 0
            help = sparse.dia_matrix((help, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            help2 = np.zeros((ncomp * nnodes))
            help2[inddir] = 1
            help2 = sparse.dia_matrix((help2, 0), shape=(ncomp * nnodes, ncomp * nnodes))
            A = help.dot(A.dot(help)) + help2.dot(A.dot(help2))
        return A.tobsr(), b, u

    # def computeBdryMean(self, u, data):
    #     colors = [int(x) for x in data.split(',')]
    #     mean, omega = np.zeros(shape=(self.ncomp,len(colors))), np.zeros(len(colors))
    #     for i,color in enumerate(colors):
    #         faces = self.mesh.bdrylabels[color]
    #         normalsS = self.mesh.normals[faces]
    #         dS = linalg.norm(normalsS, axis=1)
    #         omega[i] = np.sum(dS)
    #         for icomp in range(self.ncomp):
    #             mean[icomp,i] += np.sum(dS * np.mean(u[icomp + self.ncomp * self.mesh.faces[faces]], axis=1))
    #     return mean/omega
    #
    def computeBdryDn(self, u, data):
        colors = [int(x) for x in data.split(',')]
        flux, omega = np.zeros(shape=(len(colors),self.ncomp)), np.zeros(len(colors))
        for i,color in enumerate(colors):
            faces = self.mesh.bdrylabels[color]
            normalsS = self.mesh.normals[faces]
            dS = linalg.norm(normalsS, axis=1)
            omega[i] = np.sum(dS)
            bs, As = self.bdrydata.bsaved[color], self.bdrydata.Asaved[color]
            res = bs - As * u
            for icomp in range(self.ncomp):
                flux[i, icomp] = np.sum(res[icomp::self.ncomp])
            # else:
            #     raise NotImplementedError("computeBdryDn for condition '{}'".format(bdrycond.type[color]))
        return flux
        # colors = [int(x) for x in data.split(',')]
        # flux, omega = np.zeros(shape=(self.ncomp,len(colors))), np.zeros(len(colors))
        # for i,color in enumerate(colors):
        #     bs, As = self.bdrydata.bsaved[color], self.bdrydata.Asaved[color]
        #     res = bs - As * u
        #     for icomp in range(self.ncomp):
        #         flux[icomp, i] = np.sum(res[icomp::self.ncomp])
        # return flux
        # colors = [int(x) for x in data.split(',')]
        # omega = 0
        # for color in colors:
        #     omega += np.sum(linalg.norm(self.mesh.normals[self.mesh.bdrylabels[color]],axis=1))
        # print("###",self.bsaved[key].shape)
        # print("###",(self.Asaved[key] * u).shape)
        # res = bs - As * u
        # flux  = []
        # for icomp in range(self.ncomp):
        #     flux.append(np.sum(res[icomp::self.ncomp]))
        # return flux

    def postProcess(self, u):
        info = {}
        cell_data = {}
        point_data = {}
        for icomp in range(self.ncomp):
            point_data['U_{:02d}'.format(icomp)] = self.fem.tonode(u[icomp::self.ncomp])
        if self.problemdata.solexact:
            info['error'] = {}
            err, e = self.fem.computeErrorL2(self.problemdata.solexact, u)
            info['error']['L2'] = np.sum(err)
            for icomp in range(self.ncomp):
                point_data['E_{:02d}'.format(icomp)] = self.fem.tonode(e[icomp])
        info['postproc'] = {}
        if self.problemdata.postproc:
            for key, val in self.problemdata.postproc.items():
                type,data = val.split(":")
                if type == "bdrymean":
                    info['postproc'][key] = self.fem.computeBdryMean(u, data)
                elif type == "bdrydn":
                    info['postproc'][key] = self.computeBdryDn(u, data)
                elif type == "pointvalues":
                    info['postproc'][key] = self.fem.computePointValues(u, data)
                else:
                    raise ValueError("unknown postprocess '{}' for key '{}'".format(type, key))
        return point_data, cell_data, info

    def linearSolver(self, A, b, u=None, solver = 'umf', verbose=0):
        if not sparse.isspmatrix_bsr(A): raise ValueError("no bsr matrix")
        if solver == 'umf':
            return splinalg.spsolve(A, b, permc_spec='COLAMD'), 1
        elif solver in ['gmres','lgmres','bicgstab','cg']:
            M2 = splinalg.spilu(A, drop_tol=0.2, fill_factor=2)
            M_x = lambda x: M2.solve(x)
            M = splinalg.LinearOperator(A.shape, M_x)
            counter = tools.iterationcounter.IterationCounter(name=solver)
            args=""
            if solver == 'lgmres': args = ', inner_m=20, outer_k=4'
            cmd = "u = splinalg.{}(A, b, M=M, callback=counter {})".format(solver,args)
            exec(cmd)
            return u, counter.niter
        elif solver == 'pyamg':
            import pyamg
            config = pyamg.solver_configuration(A, verb=False)
            # ml = pyamg.smoothed_aggregation_solver(A, B=config['B'], smooth='energy')
            # ml = pyamg.smoothed_aggregation_solver(A, B=config['B'], smooth='jacobi')
            ml = pyamg.rootnode_solver(A, B=config['B'], smooth='energy')
            # print("ml", ml)
            res=[]
            # if u is not None: print("u norm", np.linalg.norm(u))
            u = ml.solve(b, x0=u, tol=1e-12, residuals=res, accel='gmres')
            if verbose: print("pyamg {:3d} ({:7.1e})".format(len(res),res[-1]/res[0]))
            return u, len(res)
        else:
            raise ValueError("unknown solve '{}'".format(solver))

        # ml = pyamg.ruge_stuben_solver(A)
        # B = np.ones((A.shape[0], 1))
        # ml = pyamg.smoothed_aggregation_solver(A, B, max_coarse=10)
        # res = []
        # # u = ml.solve(b, tol=1e-10, residuals=res)
        # u = pyamg.solve(A, b, tol=1e-10, residuals=res, verb=False,accel='cg')
        # for i, r in enumerate(res):
        #     print("{:2d} {:8.2e}".format(i,r))
        # lu = umfpack.splu(A)
        # u = umfpack.spsolve(A, b)

#=================================================================#
if __name__ == '__main__':
    print("Pas encore de test")



def _check1setinother_(set1, set2, name1="set1", name2="set2"):
    notin2 = set1.difference(set2)
    if notin2:
        raise KeyError(f"in '{name1}' but not in '{name2}': '{notin2}' {set1=} {set2=}")
def _check2setsequal_(set1, set2, name1="set1", name2="set2"):
    _check1setinother_(set1, set2, name1, name2)
    _check1setinother_(set2, set1, name2, name1)

# ---------------------------------------------------------------- #
class BoundaryConditions(object):
    """
    Information on boundary conditions
    type: dictionary int->string
    fct: dictionary int->callable
    param: dictionary int->float
    Information can be set with the 'set' function
    """
    def __init__(self, colors=None):
        if colors is None:
            self.type = {}
            self.fct = {}
            self.param = {}
        else:
            self.type = {color: None for color in colors}
            self.fct = {color: None for color in colors}
            self.param = {color: None for color in colors}
    def __repr__(self):
        return f"types={self.type}\nfct={self.fct}\nparam={self.param}"
    def clear(self):
        self.type = {}
        self.fct = {}
        self.param = {}
    # def hasExactSolution(self):
    #     return hasattr(self, 'fctexact')
    def colors(self):
        return self.type.keys()
    def types(self):
        return self.type.values()
    def set(self, type, colors, fcts=None):
        if isinstance(colors, int): colors = [colors]
        for i,color in enumerate(colors):
            self.type[color] = type
            if fcts: self.fct[color] = fcts[i]
    def colorsOfType(self, type):
        colors = []
        for color, typeofcolor in self.type.items():
            if typeofcolor == type: colors.append(color)
        return colors
    def check(self, colors):
        colors = set(colors)
        typecolors = set(self.type.keys())
        _check2setsequal_(colors, typecolors, "mesh colors", "types")


# ---------------------------------------------------------------- #
class PostProcess(object):
    """
    Information on postprocess
    type: dictionary string(name)->string
    color: dictionary string(name)->list(int)
    """
    def __init__(self):
        self.type = {}
        self.color = {}
    def __repr__(self):
        return f"types={self.type}\ncolor={self.color}"
    def clear(self):
        self.type = {}
        self.color = {}
    def set(self, name, type, colors):
        colors = list(colors)
        self.type[name] = type
        self.color[name] = colors
    def colors(self, name):
        return self.color[name]
    def check(self, colors):
        if self.type.keys() != self.color.keys():
            raise KeyError(f"postprocess keys differ: type={self.type.keys()}, color={self.color.keys()}")
        colors = set(colors)
        usedcolors = set().union(*self.color.values())
        _check1setinother_(usedcolors, colors, "used", "mesh colors")
    def colorsOfType(self, type):
        colors = []
        for n,t in self.type.items():
            if t == type: colors.extend(self.color[n])
        return colors

# ---------------------------------------------------------------- #
class Params(object):
    """
    Holds all parameters for a problem:
    - fct_glob: dictionary name -> function
    - scal_glob: dictionary name -> float
    - scal_cells dictionary name -> color -> float
    """
    def __init__(self):
        self.fct_glob = {}
        self.scal_glob = {}
        self.scal_cells = {}
    def __repr__(self):
        repr = ""
        if len(self.fct_glob): repr += f"fct_glob={self.fct_glob}"
        if len(self.scal_glob): repr += f"scal_glob={self.scal_glob}"
        if len(self.scal_cells): repr += f"scal_cells={self.scal_cells}"
        return repr
    def set_scal_cells(self, name, colors, value):
        if not name in self.scal_cells: self.scal_cells[name]={}
        for color in colors: self.scal_cells[name][color] = value
    def check(self, mesh):
        for name in self.scal_cells:
            _check2setsequal_(set(self.scal_cells[name]), set(mesh.cellsoflabel.keys()), "scal_cells", "mesh.cellsoflabel")
        for name in self.scal_glob:
            if name in self.scal_cells: raise ValueError(f"key '{name}' given twice")
            if name in self.fct_glob: raise ValueError(f"key '{name}' given twice")
            if not isinstance(self.scal_glob[name], (int,float)):
                raise ValueError(f"in 'scal_glob' key '{name}' doesnt have floats but is {self.scal_glob[name]}")
        for name in self.scal_cells:
            if name in self.scal_glob: raise ValueError(f"key '{name}' given twice")
            if name in self.fct_glob: raise ValueError(f"key '{name}' given twice")
        for name in self.fct_glob:
            if name in self.scal_cells: raise ValueError(f"key '{name}' given twice")
            if name in self.scal_glob: raise ValueError(f"key '{name}' given twice")
    def paramdefined(self, name):
        return name in self.scal_glob or name in self.scal_cells or name in self.fct_glob


# ---------------------------------------------------------------- #
class ProblemData(object):
    """
    Contains data for definition of a problem:
    - boundary conditions
    - right-hand sides
    - exact solution (if ever)
    - postprocess
    - params: class Params
    """
    def __init__(self, bdrycond=None, rhs=None, rhscell=None, rhspoint = None, postproc=None, ncomp=1):
        self.ncomp=ncomp
        if bdrycond is None: self.bdrycond = BoundaryConditions()
        else: self.bdrycond = bdrycond
        if postproc is None: self.postproc = PostProcess()
        else: self.postproc = postproc
        # self.rhs = None
        # self.rhscell = rhscell
        # self.rhspoint = rhspoint
        self.solexact = None
        self.params = Params()

    def _split2string(self, string, sep='\n\t\t'):
        return sep+sep.join(str(string).split('\n'))

    def __repr__(self):
        repr = f"\n{self.__class__}:"
        repr += f"\n\tncomp = {self.ncomp:2d}"
        repr += f"\n\tbdrycond:{self._split2string(self.bdrycond)}"
        repr += f"\n\tpostproc:{self._split2string(self.postproc)}"
        # if self.rhs: repr += f"\n\trhs={self.rhs}"
        # if self.rhscell: repr += f"\n\trhscell={self.rhscell}"
        # if self.rhspoint: repr += f"\n\trhspoint={self.rhspoint}"
        if self.solexact: repr += f"\n\tsolexact={self.solexact}"
        repr += f"\n\tparams:{self._split2string(self.params)}"
        return repr

    def check(self, mesh):
        colors = mesh.bdrylabels.keys()
        self.bdrycond.check(colors)
        if self.postproc: self.postproc.check(colors)
        self.params.check(mesh)

    def clear(self):
        """
        keeps the boundary condition types and parameters !
        """
        # self.rhs = None
        # self.rhscell = None
        # self.rhspoint = None
        self.solexact = None
        self.postproc = None
        for color in self.bdrycond.fct:
            self.bdrycond.fct[color] = None

# ---------------------------------------------------------------- #
class Results(object):
    """
    Contains results from an application:
    - point_data, side_data, cell_data, gobal_data
    - info on iteration
    """
    def __init__(self):
        self.data = {"point":{}, "side":{}, "cell":{}, "global":{}}
        self.info = {}
    def __repr__(self):
        return f"{self.data=}\n {self.info=}"
    def setData(self, data):
        if len(data) != 4:
            raise ValueError("expect four data (point, side, cell, global)")
        self.data["point"] = data[0]
        self.data["side"] = data[1]
        self.data["cell"] = data[2]
        self.data["global"] = data[3]


# -*- coding: utf-8  -*-
"""

"""

import numpy as np
import sympy


#=================================================================#
class AnalyticalFunction():
    """
    computes numpy vectorized functions for the function and its dericatives up to two
    for a given expression, derivatives computed with sympy
    """
    def __repr__(self):
        return f"dim={self.dim} expr={str(self.expr)}"
    def __call__(self, *x):
        return self.fct(*x)
    def __init__(self, expr, dim=3):
        if expr.find('x0') == -1 and expr.find('x1') == -1 and expr.find('x2') == -1:
            expr = expr.replace('x', 'x0')
            expr = expr.replace('y', 'x1')
            expr = expr.replace('z', 'x2')
        if dim==1 and expr.find('x0') == -1:
            expr = expr.replace('x', 'x0')
        self.dim, self.expr = dim, expr
        symbc = ""
        for i in range(dim): symbc += f"x{i},"
        symbc = symbc[:-1]
        s = sympy.symbols(symbc)
        # print(f"{expr=} {symbc=} {s=}")
        self.fct = np.vectorize(sympy.lambdify(symbc,expr))
        self.fct_x = []
        self.fct_xx = []
        for i in range(dim):
            if dim==1: fx = sympy.diff(expr, s)
            else: fx = sympy.diff(expr, s[i])
            self.fct_x.append(np.vectorize(sympy.lambdify(symbc, fx),otypes=[float]))
            self.fct_xx.append([])
            for j in range(dim):
                if dim == 1: fxx = sympy.diff(fx, s)
                else: fxx = sympy.diff(fx, s[j])
                self.fct_xx[i].append(np.vectorize(sympy.lambdify(symbc, fxx),otypes=[float]))
    def d(self, i, *x):
        return self.fct_x[i](*x)
    def x(self, *x):
        return self.fct_x[0](*x)
    def y(self, *x):
        return self.fct_x[1](*x)
    def z(self, *x):
        return self.fct_x[2](*x)
    def dd(self, i, j, *x):
        return self.fct_xx[i][j](*x)
    def xx(self, *x):
        return self.fct_xx[0][0](*x)
    def xy(self, *x):
        return self.fct_xx[0][1](*x)
    def xz(self, *x):
        return self.fct_xx[0][2](*x)
    def yy(self, *x):
        return self.fct_xx[1][1](*x)
    def yx(self, *x):
        return self.fct_xx[1][0](*x)
    def yz(self, *x):
        return self.fct_xx[1][2](*x)
    def zz(self, *x):
        return self.fct_xx[2][2](*x)
    def zx(self, *x):
        return self.fct_xx[2][0](*x)
    def zy(self, *x):
        return self.fct_xx[2][1](*x)


#=================================================================#
# class AnalyticalSolution3d():
#     """
#     computes numpy vectorized functions for the function and its dericatives up to two
#     for a given expression, derivatives computed with sympy
#     """
#     def __init__(self, expr):
#         (x, y, z) = sympy.symbols('x,y,z')
#         self.expr = expr
#         self.fct = np.vectorize(sympy.lambdify('x,y,z',expr))
#         fx = sympy.diff(expr, x)
#         fy = sympy.diff(expr, y)
#         fz = sympy.diff(expr, z)
#         fxx = sympy.diff(fx, x)
#         fxy = sympy.diff(fx, y)
#         fxz = sympy.diff(fx, z)
#         fyy = sympy.diff(fy, y)
#         fyz = sympy.diff(fy, z)
#         fzz = sympy.diff(fz, z)
#         self.fct_x = np.vectorize(sympy.lambdify('x,y,z', fx),otypes=[float])
#         self.fct_y = np.vectorize(sympy.lambdify('x,y,z', fy),otypes=[float])
#         self.fct_z = np.vectorize(sympy.lambdify('x,y,z', fz),otypes=[float])
#         self.fct_xx = np.vectorize(sympy.lambdify('x,y,z', fxx),otypes=[float])
#         self.fct_xy = np.vectorize(sympy.lambdify('x,y,z', fxy),otypes=[float])
#         self.fct_xz = np.vectorize(sympy.lambdify('x,y,z', fxz),otypes=[float])
#         self.fct_yy = np.vectorize(sympy.lambdify('x,y,z', fyy),otypes=[float])
#         self.fct_yz = np.vectorize(sympy.lambdify('x,y,z', fyz),otypes=[float])
#         self.fct_zz = np.vectorize(sympy.lambdify('x,y,z', fzz),otypes=[float])
#     def __repr__(self):
#         return str(self.expr)
#     def __call__(self, x, y=0, z=0):
#         return self.fct(x,y, z)
#     def x(self, x, y=0, z=0):
#         return self.fct_x(x,y, z)
#     def y(self, x, y, z=0):
#         return self.fct_y(x,y, z)
#     def z(self, x, y, z):
#         return self.fct_z(x,y, z)
#     def xx(self, x, y=0, z=0):
#         return self.fct_xx(x,y,z )
#     def yy(self, x, y, z=0):
#         return self.fct_yy(x,y,z)
#     def zz(self, x, y, z):
#         return self.fct_zz(x,y,z)
#     def d(self, i, x, y, z):
#         if i==2:    return self.fct_z(x,y,z)
#         elif i==1:  return self.fct_y(x,y,z)
#         return self.fct_x(x,y,z)
#     def dd(self, i, j, x, y, z=0):
#         if i==2:
#             if j==2:    return self.fct_zz(x,y,z)
#             elif j==1:  return self.fct_yz(x,y,z)
#             return self.fct_xz(x,y,z)
#         if i==1:
#             if j==2:    return self.fct_yz(x,y,z)
#             elif j==1:  return self.fct_yy(x,y,z)
#             return self.fct_xy(x,y,z)
#         if j==2:    return self.fct_xz(x,y,z)
#         elif j==1:  return self.fct_xy(x,y,z)
#         return self.fct_xx(x,y,z)

#=================================================================#
def analyticalSolution(function, dim, ncomp=1, random=True):
    """
    defines some analytical functions to be used in validation

    returns analytical function (if ncomp==1) or list of analytical functions (if ncomp>1)

    parameters:
        function: name of function
        ncomp: size of list
        random: use random coefficients
    """
    solexact = []
    from itertools import permutations
    def _p(i, n):
        if random:
            p = (4 * np.random.rand(n) - 2) / 3
        else:
            p = [1.1 * (4 - d) for d in range(n)]
        perm = list(permutations(p))
        return perm[i % ncomp]
    vars = ['x', 'y', 'z']
    for i in range(ncomp):
        fct = '{:3.1f}'.format(_p(i,1)[0])
        if function == 'Linear' or function == 'Quadratic':
            p = _p(i, dim)
            for d in range(dim): fct += "+{:3.1f}*{:1s}".format(p[d], vars[d])
            if function == 'Quadratic':
                p = _p(i, dim)
                for d in range(dim): fct += "+{:3.1f}*{:1s}**2".format(p[d], vars[d])
        elif function == 'Sinus':
            p = _p(i, dim)
            for d in range(dim): fct += "+{:3.1f}*sin({:1s})".format(p[d], vars[d])
        else:
            fct = function
        solexact.append(AnalyticalFunction(expr=fct))
    if ncomp==1: return solexact[0]
    return solexact


# ------------------------------------------------------------------- #
if __name__ == '__main__':
    def test1D():
        u = AnalyticalFunction(dim=1, expr='x*x')
        print("u(2)", u(2))
        x = np.meshgrid(np.linspace(0, 2, 3))
        print("x", x, "\nu=", u.expr, "\nu(x)", u(x), "\nu.x(x)", u.x(x), "\nu.xx(x)", u.xx(x))
    def test2D():
        u = AnalyticalFunction(dim=2, expr='x*x*y + y*y')
        print("u(2,1)", u(2,1))
        print("u(2,1)", u(*(2,1)))
        x = np.meshgrid(np.linspace(0, 2, 3),np.linspace(0, 1, 2))
        print("x", x, "\nu=", u.expr, "\nu(x)", u(*x), "\nu.x(x)", u.x(*x), "\nu.xx(x)", u.xx(*x))
    # test2D()
    test1D()






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 10:49:13 2020

@author: becker
"""

import numpy as np
import scipy

k = 3
d = 1

def sums(length, total_sum):
    if length == 1:
        yield (total_sum,)
    else:
        for value in range(total_sum + 1):
            for permutation in sums(length - 1, total_sum - value):
                yield (value,) + permutation

for i in sums(d+1,k):
    print(f"{i=} {scipy.special.factorial(np.array(i, dtype=int))}")

L = list(sums(d+1,k))
print('total permutations:',len(L))
print(f"{L=}")



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""

import numpy as np
import scipy.sparse as spsp
import scipy.sparse.linalg as splinalg
# from functools import partial

import simfempy.tools.analyticalfunction
import simfempy.tools.timer
import simfempy.tools.iterationcounter
import simfempy.applications.problemdata

#=================================================================#
class Application(object):
    def __init__(self, **kwargs):
        self.linearsolvers=['umf', 'lgmres', 'bicgstab']
        try:
            import pyamg
            self.linearsolvers.append('pyamg')
        except:
            import warnings
            warnings.warn("*** pyamg not found (umf used instead)***")
        self.linearsolver = 'umf'
        self.timer = simfempy.tools.timer.Timer(verbose=0)
        if 'problemdata' in kwargs:
            # self.problemdata = copy.deepcopy(kwargs.pop('problemdata'))
            self.problemdata = kwargs.pop('problemdata')
            self.ncomp = self.problemdata.ncomp
            assert self.ncomp != -1
        if 'exactsolution' in kwargs:
            self.exactsolution = kwargs.pop('exactsolution')
            self._generatePDforES = True
            if 'random' in kwargs:
                self.random_exactsolution = kwargs.pop('random')
            else:
                self.random_exactsolution = False
        else:
            self._generatePDforES = False
        if 'geom' in kwargs:
            self.geom = kwargs.pop('geom')
            print(f"{self.geom=}")
            mesh = self.geom.generate_mesh()
            self.mesh = simfempy.meshes.simplexmesh.SimplexMesh(mesh=mesh)
        if 'mesh' in kwargs:
            self.mesh = kwargs.pop('mesh')
        self._setMeshCalled = False
    def setParameter(self, paramname, param):
        assert 0
    def dirichletfct(self):
        if self.ncomp > 1:
            def _solexactdir(x, y, z):
                return [self.problemdata.solexact[icomp](x, y, z) for icomp in range(self.ncomp)]
        else:
            def _solexactdir(x, y, z):
                return self.problemdata.solexact(x, y, z)
        return _solexactdir
    def generatePoblemDataForAnalyticalSolution(self):
        bdrycond = self.problemdata.bdrycond
        self.problemdata.solexact = self.defineAnalyticalSolution(exactsolution=self.exactsolution, random=self.random_exactsolution)
        print("self.problemdata.solexact", self.problemdata.solexact)
        self.problemdata.params.fct_glob['rhs'] = self.defineRhsAnalyticalSolution(self.problemdata.solexact)
        for color in self.mesh.bdrylabels:
            if color in bdrycond.type and bdrycond.type[color] in ["Dirichlet","dirichlet"]:
                bdrycond.fct[color] = self.dirichletfct()
            else:
                if color in bdrycond.type:
                    cmd = "self.define{}AnalyticalSolution(self.problemdata,{})".format(bdrycond.type[color], color)
                    # print(f"cmd={cmd}")
                    bdrycond.fct[color] = eval(cmd)
                else:
                    bdrycond.fct[color] = self.defineBdryFctAnalyticalSolution(color)
        # for color in self.mesh.bdrylabels:
        #     if not color in bdrycond.type: raise KeyError(f"{color=} {bdrycond.type=}")
        #     if bdrycond.type[color] in ["Dirichlet"]:
        #     else:
        #         type = bdrycond.type[color]
        #         cmd = "self.define{}AnalyticalFunction(self.problemdata,{})".format(type, color)
        #         # print(f"cmd={cmd}")
        #         bdrycond.fct[color] = eval(cmd)
    def defineAnalyticalSolution(self, exactsolution, random=True):
        dim = self.mesh.dimension
        return simfempy.tools.analyticalfunction.analyticalSolution(exactsolution, dim, self.ncomp, random)
    def setMesh(self, mesh):
        self.mesh = mesh
        self._setMeshCalled = True
        if hasattr(self,'_generatePDforES') and self._generatePDforES:
            self.generatePoblemDataForAnalyticalSolution()
            self._generatePDforES = False
    def compute_cell_vector_from_params(self, name, params):
        if name in params.fct_glob:
            xc, yc, zc = self.mesh.pointsc.T
            fct = np.vectorize(params.fct_glob[name])
            arr = fct(self.mesh.cell_labels, xc, yc, zc)
        elif name in params.scal_glob:
            arr = np.full(self.mesh.ncells, params.scal_glob[name])
        elif name in params.scal_cells:
            arr = np.empty(self.mesh.ncells)
            for color in params.scal_cells[name]:
                arr[self.mesh.cellsoflabel[color]] = params.scal_cells[name][color]
        else:
            msg = f"{name} should be given in 'fct_glob' or 'scal_glob' or 'scal_cells' (problemdata.params)"
            raise ValueError(msg)
        return arr


    def static(self, iter=100, dirname='Run'):
        # print(f"### static")
        if not self._setMeshCalled: self.setMesh(self.mesh)
        # self.timer.reset_all()
        return self.solveLinearProblem()

    # def solve(self, iter=0, dirname="Results"):
    #     return self.solveLinearProblem()

    def solveLinearProblem(self):
        # print(f"### solveLinearProblem")
        if not hasattr(self,'mesh'): raise ValueError("*** no mesh given ***")
        result = simfempy.applications.problemdata.Results()
        self.timer.add('init')
        A = self.computeMatrix()
        self.timer.add('matrix')
        b, u = self.computeRhs()
        self.timer.add('rhs')
        u, niter = self.linearSolver(A, b, u, solver=self.linearsolver)
        # print(f"{u=}")
        self.timer.add('solve')
        result.setData(self.postProcess(u))
        self.timer.add('postp')
        result.info['timer'] = self.timer
        result.info['iter'] = {'lin':niter}
        return result


    # def solveNonlinearProblem(self, u=None, sdata=None, method="newton", checkmaxiter=True):
    #     if not hasattr(self,'mesh'): raise ValueError("*** no mesh given ***")
    #     self.timer.add('init')
    #     A = self.matrix()
    #     self.timer.add('matrix')
    #     self.b,u = self.computeRhs(u)
    #     self.du = np.empty_like(u)
    #     self.timer.add('rhs')
    #     if method == 'newton':
    #         from . import newton
    #         u, nit = newton.newton(x0=u, f=self.residualNewton, computedx=self.solveForNewton, sdata=sdata, verbose=True)
    #     elif method in ['broyden2','krylov', 'df-sane', 'anderson']:
    #         sol = optimize.root(self.residualNewton, u, method=method)
    #         u, nit = sol.x, sol.nit
    #     else:
    #         raise ValueError(f"unknown method {method}")
    #     point_data, cell_data, info = self.postProcess(u)
    #     self.timer.add('postp')
    #     info['timer'] = self.timer
    #     info['iter'] = {'lin':nit}
    #     return point_data, cell_data, info

    # def residualNewton(self, u):
    #     # print(f"self.b={self.b.shape}")
    #     self.A = self.matrix()
    #     return self.A.dot(u) - self.b
    #
    # def solveForNewton(self, r, x):
    #     # print(f"solveForNewton r={np.linalg.norm(r)}")
    #     du, niter = self.linearSolver(self.A, r, x, verbose=0)
    #     # print(f"solveForNewton du={np.linalg.norm(du)}")
    #     return du

    def linearSolver(self, A, b, u=None, solver = None, verbose=1):
        print(f"### linearSolver {solver=}")
        if len(b.shape)!=1 or len(A.shape)!=2 or b.shape[0] != A.shape[0]:
            raise ValueError(f"{A.shape=} {b.shape=}")
        if solver is None: solver = self.linearsolver
        if not hasattr(self, 'info'): self.info={}
        if solver not in self.linearsolvers: solver = "umf"
        if solver == 'umf':
            return splinalg.spsolve(A, b, permc_spec='COLAMD'), 1
        elif solver in ['gmres','lgmres','bicgstab','cg']:
            if solver == 'cg':
                def gaussSeidel(A):
                    dd = A.diagonal()
                    D = spsp.dia_matrix(A.shape)
                    D.setdiag(dd)
                    L = spsp.tril(A, -1)
                    U = spsp.triu(A, 1)
                    return splinalg.factorized(D + L)
                M2 = gaussSeidel(A)
            else:
                # defaults: drop_tol=0.0001, fill_factor=10
                M2 = splinalg.spilu(A.tocsc(), drop_tol=0.1, fill_factor=3)
            M_x = lambda x: M2.solve(x)
            M = splinalg.LinearOperator(A.shape, M_x)
            counter = simfempy.tools.iterationcounter.IterationCounter(name=solver, verbose=verbose)
            args=""
            cmd = "u = splinalg.{}(A, b, M=M, tol=1e-14, callback=counter {})".format(solver,args)
            exec(cmd)
            return u, counter.niter
        elif solver == 'pyamg':
            import pyamg
            res=[]
            # u = pyamg.solve(A=A, b=b, x0=u, tol=1e-14, residuals=res, verb=False)
            B = np.ones((A.shape[0], 1))
            SA_build_args = {
                'max_levels': 10,
                'max_coarse': 25,
                'coarse_solver': 'pinv2',
                'symmetry': 'hermitian'}
            smooth = ('energy', {'krylov': 'cg'})
            strength = [('evolution', {'k': 2, 'epsilon': 4.0})]
            presmoother = ('gauss_seidel', {'sweep': 'symmetric', 'iterations': 1})
            postsmoother = ('gauss_seidel', {'sweep': 'symmetric', 'iterations': 1})
            strength = [('evolution', {'k': 2, 'epsilon': 4.0})]
            presmoother = ('gauss_seidel', {'sweep': 'symmetric', 'iterations': 1})
            postsmoother = ('gauss_seidel', {'sweep': 'symmetric', 'iterations': 1})
            ml = pyamg.smoothed_aggregation_solver(A, B, max_coarse=10)
            # ml = pyamg.smoothed_aggregation_solver(
            #     A,
            #     B=B,
            #     smooth=smooth,
            #     strength=strength,
            #     presmoother=presmoother,
            #     postsmoother=postsmoother,
            #     **SA_build_args)
            # u= ml.solve(b=b, x0=u, tol=1e-12, residuals=res, **SA_solve_args)
            SA_solve_args = {'cycle': 'V', 'maxiter': 50, 'tol': 1e-14}
            u= ml.solve(b=b, x0=u, residuals=res, **SA_solve_args)
            if(verbose): print('niter ({}) {:4d} ({:7.1e})'.format(solver, len(res),res[-1]/res[0]))
            return u, len(res)
        else:
            raise NotImplementedError("unknown solve '{}'".format(solver))



# ------------------------------------- #

if __name__ == '__main__':
    raise ValueError("unit test to be written")


# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016

@author: becker
"""
import os, sys, importlib
import meshio
import numpy as np
from scipy import sparse
from simfempy.tools import npext
from simfempy.tools import timer

from .testmeshes import __pygmsh6__

#=================================================================#
class SimplexMesh(object):
    """
    simplicial mesh, can be initialized from the output of pygmsh.
    Needs physical labels geometry objects of highest dimension and co-dimension one

    dimension, nnodes, ncells, nfaces: dimension, number of nodes, simplices, faces
    points: coordinates of the vertices of shape (nnodes,3)
    pointsc: coordinates of the barycenters of cells (ncells,3)
    pointsf: coordinates of the barycenters of faces (nfaces,3)

    simplices: node ids of simplices of shape (ncells, dimension+1)
    faces: node ids of faces of shape (nfaces, dimension)

    facesOfCells: shape (ncells, dimension+1): contains simplices[i,:]-setminus simplices[i,ii], sorted
    cellsOfFaces: shape (nfaces, 2): cellsOfFaces[i,1]=-1 if boundary

    normals: normal per face of length dS, oriented from  ids of faces of shape (nfaces, dimension)
             normals on boundary are external
    sigma: orientation of normal per cell and face (ncells, dimension+1)

    dV: shape (ncells), volumes of simplices
    bdrylabels: dictionary(keys: colors, values: id's of boundary faces)
    cellsoflabel: dictionary(keys: colors, values: id's of cells)
    """

    def __repr__(self):
        s = f"SimplexMesh({self.geometry}): "
        s += f"dim/nnodes/nfaces/ncells: {self.dimension}/{self.nnodes}/{self.nfaces}/{self.ncells}"
        s += f"\nbdrylabels={list(self.bdrylabels.keys())}"
        s += f"\ncellsoflabel={list(self.cellsoflabel.keys())}"
        return s

    def __init__(self, **kwargs):
        if 'mesh' in kwargs:
            self.geometry = 'own'
            mesh = kwargs.pop('mesh')
        else:
            raise KeyError("Needs a mesh (no longer geometry)")
        self._initMeshPyGmsh(mesh)
    def _initMeshPyGmsh(self, mesh):
        # only 3d-coordinates
        assert mesh.points.shape[1] ==3
        self.points = mesh.points
        self.nnodes = self.points.shape[0]
        self.celltypes = [key for key, cellblock in mesh.cells]
        # for key, cellblock in cells: keys.append(key)
        # print("self.celltypes", self.celltypes)
        if 'tetra' in self.celltypes:
            self.dimension = 3
            self.simplicesname, self.facesname = 'tetra', 'triangle'
        elif 'triangle' in self.celltypes:
            self.dimension = 2
            self.simplicesname, self.facesname = 'triangle', 'line'
        elif 'line' in self.celltypes:
            self.dimension = 1
            self.simplicesname, self.facesname = 'line', 'vertex'
        else:
            raise ValueError(f"something wrong {self.celltypes=} {mesh=}")
        bdryfacesgmshlist = []
        for key, cellblock in mesh.cells:
            # print(f"{key=} {cellblock=}")
            if key == self.simplicesname: self.simplices = cellblock
            if key == 'vertex': self.vertices = cellblock
            if key == self.facesname:
                # print(f"{key=} {cellblock=}")
                bdryfacesgmshlist.extend(cellblock)
        if not hasattr(self,"simplices"):
            raise ValueError(f"something wrong {self.dimension=}")
        bdryfacesgmsh = np.array(bdryfacesgmshlist)
        self._constructFacesFromSimplices()
        assert self.dimension+1 == self.simplices.shape[1]
        self.ncells = self.simplices.shape[0]
        self.pointsc = self.points[self.simplices].mean(axis=1)
        self.pointsf = self.points[self.faces].mean(axis=1)
        self._constructNormalsAndAreas()
        if __pygmsh6__:
            self._initMeshPyGmsh6(mesh.cells, mesh.cell_data['gmsh:physical'], bdryfacesgmsh)
        else:
            self._initMeshPyGmsh7(mesh.cells, mesh.cell_sets, bdryfacesgmsh)

    def _initMeshPyGmsh7(self, cells, cell_sets, bdryfacesgmsh):
        # cell_sets: dict label --> list of None or np.array for each cell_type
        # the indices of the np.array are not the cellids !
        # ???
        # print(f"{cell_sets=}")
        typesoflabel = {}
        sizes = {key:0 for key in self.celltypes}
        cellsoflabel = {key:{} for key in self.celltypes}
        ctorderd = []
        for label, cb in cell_sets.items():
            # print(f"{label=} {cb=}")
            if len(cb) != len(self.celltypes): raise KeyError(f"mismatch {label=}")
            for celltype, info in zip(self.celltypes, cb):
                # only one is supposed to be not None
                if info is not None:
                    # print(f"{label=} {type(label)=}")
                    try: ilabel=int(label)
                    except: raise ValueError(f"cannot convert to int {label=} {cell_sets=}")
                    cellsoflabel[celltype][ilabel] = info
                    sizes[celltype] += info.shape[0]
                    typesoflabel[ilabel] = celltype
                    ctorderd.append(celltype)
        #correcting the numbering in cell_sets
        n = 0
        for ct in list(dict.fromkeys(ctorderd)):
            #eliminates duplicates
            for l, cb in cellsoflabel[ct].items(): cb -= n
            n += sizes[ct]
        self.cellsoflabel = cellsoflabel[self.simplicesname]
        self.verticesoflabel = {}
        if self.dimension > 1: self.verticesoflabel = cellsoflabel['vertex']
        # print(f"{self.verticesoflabel=}")
        # bdry faces
        # for key, cellblock in cells:
        #     if key == self.facesnames[self.dimension - 1]: bdryfacesgmsh = cellblock
        bdrylabelsgmsh = cellsoflabel[self.facesname]
        self._constructBoundaryFaces7(bdryfacesgmsh, bdrylabelsgmsh)

    def _initMeshPyGmsh6(self, cells, cdphys, bdryfacesgmsh):
        if len(cdphys) != len(self.celltypes):
            raise KeyError(f"not enough physical labels:\n {self.celltypes=}\n {len(cdphys)=}")
        # first attempt, 'np.append' copies data...
        _cells = {}
        _labels = {}
        for (key, cellblock), cd in zip(cells,cdphys):
            if len(cellblock) != len(cd):
                raise ValueError(f"mismatch in {key} {len(cellblock)} {len(cd)}")
            if not key in _cells.keys():
                _cells[key] = cellblock
                _labels[key] = cd
            else:
                _cells[key] = np.append(_cells[key], cellblock, axis=0)
                _labels[key] = np.append(_labels[key], cd, axis=0)
        self._constructBoundaryFaces6(bdryfacesgmsh, _labels[self.facesname])
        # self.cellsoflabel = npext.creatdict_unique_all(self.cell_labels)
        self.cellsoflabel = npext.creatdict_unique_all(_labels[self.simplicesname])
        self.verticesoflabel = {}
        if self.dimension > 1 and 'vertex' in _cells.keys():
            self.vertices = _cells['vertex'].reshape(-1)
            self.verticesoflabel = npext.creatdict_unique_all(_labels['vertex'])

    def _constructFacesFromSimplices(self):
        simplices = self.simplices
        ncells = simplices.shape[0]
        nnpc = simplices.shape[1]
        allfaces = np.empty(shape=(nnpc*ncells,nnpc-1), dtype=int)
        for i in range(ncells):
            for ii in range(nnpc):
                mask = np.array( [jj !=ii for jj in range(nnpc)] )
                allfaces[i*nnpc+ii] = np.sort(simplices[i,mask])
        s = "{0}" + (nnpc-2)*", {0}"
        s = s.format(allfaces.dtype)
        order = ["f0"]+["f{:1d}".format(i) for i in range(1,nnpc-1)]
        if self.dimension==1:
            perm = np.argsort(allfaces, axis=0).ravel()
        else:
            perm = np.argsort(allfaces.view(s), order=order, axis=0).ravel()
        allfacescorted = allfaces[perm]
        self.faces, indices = np.unique(allfacescorted, return_inverse=True, axis=0)
        locindex = np.tile(np.arange(0,nnpc), ncells)
        cellindex = np.repeat(np.arange(0,ncells), nnpc)
        self.nfaces = self.faces.shape[0]
        self.cellsOfFaces = -1 * np.ones(shape=(self.nfaces, 2), dtype=int)
        self.facesOfCells = np.zeros(shape=(ncells, nnpc), dtype=int)
        for ii in range(indices.shape[0]):
            f = indices[ii]
            loc = locindex[perm[ii]]
            cell = cellindex[perm[ii]]
            self.facesOfCells[cell, loc] = f
            if self.cellsOfFaces[f,0] == -1: self.cellsOfFaces[f,0] = cell
            else: self.cellsOfFaces[f,1] = cell
    def _constructBoundaryFaces7(self, bdryfacesgmsh, bdrylabelsgmsh):
        # bdries
        # bdryfacesgmsh may contains interior edges for len(celllabels)>1
        bdryfacesgmsh = np.sort(bdryfacesgmsh)
        bdryids = np.flatnonzero(self.cellsOfFaces[:,1] == -1)
        # print(f"{bdryids=}")
        # assert np.all(bdryids == np.flatnonzero(np.any(self.cellsOfFaces == -1, axis=1)))
        bdryfaces = np.sort(self.faces[bdryids],axis=1)
        # print(f"{bdryfacesgmsh=}\n{bdryfaces=}")
        # ind = np.isin(bdryfacesgmsh, bdryfaces)
        # print(f"{ind=} {bdryfacesgmsh[ind]=}")
        # print(f"{bdryfaces=}")
        nbdryfaces = len(bdryids)
        nnpc = self.simplices.shape[1]
        s = "{0}" + (nnpc-2)*", {0}"
        dtb = s.format(bdryfacesgmsh.dtype)
        dtf = s.format(bdryfaces.dtype)
        order = ["f0"]+["f{:1d}".format(i) for i in range(1,nnpc-1)]
        if self.dimension==1:
            bp = np.argsort(bdryfacesgmsh.view(dtb), axis=0).ravel()
            fp = np.argsort(bdryfaces.view(dtf), axis=0).ravel()
        else:
            bp = np.argsort(bdryfacesgmsh.view(dtb), order=order, axis=0).ravel()
            fp = np.argsort(bdryfaces.view(dtf), order=order, axis=0).ravel()
        # print(f"{bp=}")
        # print(f"{fp=}")

#https://stackoverflow.com/questions/51352527/check-for-identical-rows-in-different-numpy-arrays
        indices = (bdryfacesgmsh[bp, None] == bdryfaces[fp]).all(-1).any(-1)
        # indices = np.isin(bdryfacesgmsh, bdryfaces)
        # print(f"{indices=}")

        # fp2 = np.searchsorted(bdryfacesgmsh, bdryfaces, sorter=bp)
        # print(f"{fp2=}")

        # print(f"{bp[indices]=}")
        # print(f"{bdryfacesgmsh[bp[indices]]=}")
        # print(f"{bdryfaces[fp]=}")
        if not np.all(bdryfaces[fp]==bdryfacesgmsh[bp[indices]]):
            raise ValueError(f"{bdryfaces.T=}\n{bdryfacesgmsh.T=}\n{indices=}\n{bdryfaces[fp].T=}\n{bdryfacesgmsh[bp[indices]].T=}")

        bp2 = bp[indices]
        for i in range(len(fp)):
            if not np.all(bdryfacesgmsh[bp2[i]] == bdryfaces[fp[i]]):
                raise ValueError(f"{i=} {bdryfacesgmsh[bp2[i]]=} {bdryfaces[fp[i]]=}")

        bpi = np.argsort(bp)
        # bp2i = {bp2[i]:i for i in range(len(bp2))}
        # print(f"{bp=} \n{bp2=} \n{bpi=} \n{bp2i=} \n{indices=}")
        binv = -1*np.ones_like(bp)
        binv[bp2] = np.arange(len(bp2))
        self.bdrylabels = {}
        for col, cb in bdrylabelsgmsh.items():
            # if cb[0] in bp2i.keys():
            if indices[bpi[cb[0]]]:
                # for i in range(len(cb)):
                #     if not bp2i[cb[i]] == binv[cb[i]]:
                #         raise ValueError(f"{bp2i[cb[i]]} {binv[cb[i]]}")
                # print(f"{col=}")
                self.bdrylabels[int(col)] = np.empty_like(cb)
                # for i in range(len(cb)): self.bdrylabels[int(col)][i] = bdryids[fp[bp2i[cb[i]]]]
                for i in range(len(cb)): self.bdrylabels[int(col)][i] = bdryids[fp[binv[cb[i]]]]
            # else:
            #     assert not indices[bpi[cb[0]]]



        # nbdrylabelsgmsh = 0
        # for col, cb in bdrylabelsgmsh.items(): nbdrylabelsgmsh += cb.shape[0]
        # print(f"{nbdrylabelsgmsh=} {nbdryfaces=}")
        # # if nbdrylabelsgmsh != nbdryfaces:
        # #     raise ValueError(f"wrong number of boundary labels {nbdrylabelsgmsh=} != {nbdryfaces=}")
        # if len(bdryfacesgmsh) != nbdryfaces:
        #     print("wrong number of bdryfaces {} != {}".format(len(bdryfacesgmsh), nbdryfaces))
        #     # raise ValueError("wrong number of bdryfaces {} != {}".format(len(bdryfacesgmsh), nbdryfaces))
        # self.bdrylabels = {}
        # # colorofbdr = -np.ones(nbdryfaces, dtype=bdryids.dtype)
        # colorofbdr = -np.ones(nbdrylabelsgmsh, dtype=bdryids.dtype)
        # for col, cb in bdrylabelsgmsh.items():
        #     self.bdrylabels[int(col)] = -np.ones( (cb.shape[0]), dtype=np.int32)
        #     colorofbdr[cb] = col
        #
        # # perm = bdryids[bp[fpi]]
        # # print(f"{perm=}")
        # # print(f"{colorofbdr=}")
        # counts = {}
        # for key in list(self.bdrylabels.keys()): counts[key]=0
        # for i in range(len(perm)):
        #     if np.any(bdryfacesgmsh[i] != self.faces[perm[i]]):
        #         raise ValueError(f"Did not find boundary indices\n{bdryfacesgmsh[i]} {self.faces[perm[i]]}")
        #     color = colorofbdr[i]
        #     self.bdrylabels[color][counts[color]] = perm[i]
        #     counts[color] += 1
        # # print ("self.bdrylabels", self.bdrylabels)
        # for col, bl in self.bdrylabels.items():
        #     print(f"{bl=} {self.bdrylabels2[col]=}")
        #     # assert np.all(bl == self.bdrylabels2[col])


    def _constructBoundaryFaces6(self, bdryfacesgmsh, bdrylabelsgmsh):
        # bdries
        bdryids = np.flatnonzero(self.cellsOfFaces[:,1] == -1)
        assert np.all(bdryids == np.flatnonzero(np.any(self.cellsOfFaces == -1, axis=1)))
        bdryfaces = np.sort(self.faces[bdryids],axis=1)
        nbdryfaces = len(bdryids)
        if len(bdrylabelsgmsh) != nbdryfaces:
            raise ValueError("wrong number of boundary labels {} != {}".format(len(bdrylabelsgmsh),nbdryfaces))
        if len(bdryfacesgmsh) != nbdryfaces:
            raise ValueError("wrong number of bdryfaces {} != {}".format(len(bdryfacesgmsh), nbdryfaces))
        self.bdrylabels = {}
        colors, counts = np.unique(bdrylabelsgmsh, return_counts=True)
        # print ("colors, counts", colors, counts)
        for i in range(len(colors)):
            self.bdrylabels[colors[i]] = -np.ones( (counts[i]), dtype=np.int32)
        bdryfacesgmsh = np.sort(bdryfacesgmsh)
        nnpc = self.simplices.shape[1]
        s = "{0}" + (nnpc-2)*", {0}"
        dtb = s.format(bdryfacesgmsh.dtype)
        dtf = s.format(bdryfaces.dtype)
        order = ["f0"]+["f{:1d}".format(i) for i in range(1,nnpc-1)]
        if self.dimension==1:
            bp = np.argsort(bdryfacesgmsh.view(dtb), axis=0).ravel()
            fp = np.argsort(bdryfaces.view(dtf), axis=0).ravel()
        else:
            bp = np.argsort(bdryfacesgmsh.view(dtb), order=order, axis=0).ravel()
            fp = np.argsort(bdryfaces.view(dtf), order=order, axis=0).ravel()
        bpi = np.empty(bp.size, bp.dtype)
        bpi[bp] = np.arange(bp.size)
        perm = bdryids[fp[bpi]]
        counts = {}
        for key in list(self.bdrylabels.keys()): counts[key]=0
        for i in range(len(perm)):
            if np.any(bdryfacesgmsh[i] != self.faces[perm[i]]):
                raise ValueError("Did not find boundary indices")
            color = bdrylabelsgmsh[i]
            self.bdrylabels[color][counts[color]] = perm[i]
            counts[color] += 1
        # print ("self.bdrylabels", self.bdrylabels)

    def _constructNormalsAndAreas(self):
        elem = self.simplices
        self.sigma = np.array([2 * (self.cellsOfFaces[self.facesOfCells[ic, :], 0] == ic)-1 for ic in range(self.ncells)])
        if self.dimension==1:
            x = self.points[:,0]
            self.normals = np.stack((np.ones(self.nfaces), np.zeros(self.nfaces), np.zeros(self.nfaces)), axis=-1)
            dx1 = x[elem[:, 1]] - x[elem[:, 0]]
            self.dV = np.abs(dx1)
        elif self.dimension==2:
            x,y = self.points[:,0], self.points[:,1]
            sidesx = x[self.faces[:, 1]] - x[self.faces[:, 0]]
            sidesy = y[self.faces[:, 1]] - y[self.faces[:, 0]]
            self.normals = np.stack((-sidesy, sidesx, np.zeros(self.nfaces)), axis=-1)
            dx1 = x[elem[:, 1]] - x[elem[:, 0]]
            dx2 = x[elem[:, 2]] - x[elem[:, 0]]
            dy1 = y[elem[:, 1]] - y[elem[:, 0]]
            dy2 = y[elem[:, 2]] - y[elem[:, 0]]
            self.dV = 0.5 * np.abs(dx1*dy2-dx2*dy1)
        else:
            x, y, z = self.points[:, 0], self.points[:, 1], self.points[:, 2]
            x1 = x[self.faces[:, 1]] - x[self.faces[:, 0]]
            y1 = y[self.faces[:, 1]] - y[self.faces[:, 0]]
            z1 = z[self.faces[:, 1]] - z[self.faces[:, 0]]
            x2 = x[self.faces[:, 2]] - x[self.faces[:, 0]]
            y2 = y[self.faces[:, 2]] - y[self.faces[:, 0]]
            z2 = z[self.faces[:, 2]] - z[self.faces[:, 0]]
            sidesx = y1*z2 - y2*z1
            sidesy = x2*z1 - x1*z2
            sidesz = x1*y2 - x2*y1
            self.normals = 0.5*np.stack((sidesx, sidesy, sidesz), axis=-1)
            dx1 = x[elem[:, 1]] - x[elem[:, 0]]
            dx2 = x[elem[:, 2]] - x[elem[:, 0]]
            dx3 = x[elem[:, 3]] - x[elem[:, 0]]
            dy1 = y[elem[:, 1]] - y[elem[:, 0]]
            dy2 = y[elem[:, 2]] - y[elem[:, 0]]
            dy3 = y[elem[:, 3]] - y[elem[:, 0]]
            dz1 = z[elem[:, 1]] - z[elem[:, 0]]
            dz2 = z[elem[:, 2]] - z[elem[:, 0]]
            dz3 = z[elem[:, 3]] - z[elem[:, 0]]
            self.dV = (1/6) * np.abs(dx1*(dy2*dz3-dy3*dz2) - dx2*(dy1*dz3-dy3*dz1) + dx3*(dy1*dz2-dy2*dz1))
        for i in range(self.nfaces):
            i0, i1 = self.cellsOfFaces[i, 0], self.cellsOfFaces[i, 1]
            if i1 == -1:
                xt = np.mean(self.points[self.faces[i]], axis=0) - np.mean(self.points[self.simplices[i0]], axis=0)
                if np.dot(self.normals[i], xt)<0:  self.normals[i] *= -1
            else:
                xt = np.mean(self.points[self.simplices[i1]], axis=0) - np.mean(self.points[self.simplices[i0]], axis=0)
                if np.dot(self.normals[i], xt) < 0:  self.normals[i] *= -1
        # self.sigma = np.array([1.0 - 2.0 * (self.cellsOfFaces[self.facesOfCells[ic, :], 0] == ic) for ic in range(self.ncells)])

    def write(self, filename, dirname = None, point_data=None):
        cell_data_meshio = {}
        if hasattr(self,'vertex_labels'):
            cell_data_meshio['vertex'] = {}
            cell_data_meshio['vertex']['gmsh:physical'] = self.vertex_labels
        if self.dimension ==2:
            cells = {'triangle': self.simplices}
            cells['line'] = self._facedata[0]
            cell_data_meshio['line']={}
            cell_data_meshio['line']['gmsh:physical'] = self._facedata[1]
            cell_data_meshio['triangle']={}
            cell_data_meshio['triangle']['gmsh:physical'] = self.cell_labels
        else:
            cells = {'tetra': self.simplices}
            cells['triangle'] = self._facedata[0]
            cell_data_meshio['triangle']={}
            cell_data_meshio['triangle']['gmsh:physical'] = self._facedata[1]
            cell_data_meshio['tetra']={}
            cell_data_meshio['tetra']['gmsh:physical'] = self.cell_labels
        if dirname is not None:
            dirname = dirname + os.sep + "mesh"
            if not os.path.isdir(dirname) :
                os.makedirs(dirname)
            filename = os.path.join(dirname, filename)
        print("cell_data_meshio['line']['gmsh:physical']", cell_data_meshio['line']['gmsh:physical'])
        meshio.write_points_cells(filename=filename, points=self.points, cells=cells, point_data=point_data, cell_data=cell_data_meshio, file_format='gmsh2-ascii')

    def computeSimpOfVert(self, test=False):
        S = sparse.dok_matrix((self.nnodes, self.ncells), dtype=int)
        for ic in range(self.ncells):
            S[self.simplices[ic,:], ic] = ic+1
        S = S.tocsr()
        S.data -= 1
        self.simpOfVert = S
        if test:
            # print("S=",S)
            from . import plotmesh
            import matplotlib.pyplot as plt
            simps, xc, yc = self.simplices, self.pointsc[:,0], self.pointsc[:,1]
            meshdata =  self.x, self.y, simps, xc, yc
            plotmesh.meshWithNodesAndTriangles(meshdata)
            plt.show()

    def plot(self, **kwargs):
        from simfempy.meshes import plotmesh
        plotmesh.plotmesh(self, **kwargs)
    def plotWithBoundaries(self):
        # from . import plotmesh
        from simfempy.meshes import plotmesh
        plotmesh.meshWithBoundaries(self)
    def plotWithNumbering(self, **kwargs):
        # from . import plotmesh
        from simfempy.meshes import plotmesh
        plotmesh.plotmeshWithNumbering(self, **kwargs)
    def plotWithData(self, **kwargs):
        # from . import plotmesh
        from simfempy.meshes import plotmesh
        plotmesh.meshWithData(self, **kwargs)


#=================================================================#
if __name__ == '__main__':
    import geomdefs
    geometry = geomdefs.unitsquare.Unitsquare()
    mesh = SimplexMesh(geometry=geometry, hmean=2)
    import plotmesh
    import matplotlib.pyplot as plt
    fig, axarr = plt.subplots(2, 1, sharex='col')
    plotmesh.meshWithBoundaries(mesh, ax=axarr[0])
    plotmesh.plotmeshWithNumbering(mesh, ax=axarr[1])
    plotmesh.plotmeshWithNumbering(mesh, localnumbering=True)



import numpy as np

# ------------------------------------- #
def unique_all(a):
    """
    https://stackoverflow.com/questions/30003068/get-a-list-of-all-indices-of-repeated-elements-in-a-numpy-array
    """
    a = np.asarray(a)
    ind_s = np.argsort(a)
    a_s = a[ind_s]
    vals, ind_start = np.unique(a_s, return_index=True)
    return vals, np.split(ind_s, ind_start[1:])

def creatdict_unique_all(cl):
    clunique = unique_all(cl)
    clinv = {}
    for color, ind in zip(clunique[0], clunique[1]):
        clinv[color] = ind
    return clinv


# ------------------------------------- #
if __name__ == '__main__':
    a = np.array([1, 7, 3, 1, 6, 7, 1, 6, 1, 7])
    vals, inds = unique_all(a)
    print("a", a)
    print("vals", vals)
    print("inds", inds)
    for ind, val in zip(inds,vals):
        print("val", val, "ind", ind, "val[ind]", a[ind])




import matplotlib.pyplot as plt
from . import plotmesh1d, plotmesh2d, plotmesh3d
import meshio

#----------------------------------------------------------------#
def _getDim(meshdata):
    try:
        dim = meshdata.dimension
        meshdataismesh = True
    except:
        dim = len(meshdata)-3
        meshdataismesh = False
    return dim, meshdataismesh


#=================================================================#
def plotmesh(mesh, **kwargs):
    dim, meshdataismesh = _getDim(mesh)
    if dim == 1:
        plotmesh1d.plotmesh(mesh=mesh, **kwargs)
    elif dim == 2:
        plotmesh2d.plotmesh(mesh=mesh, **kwargs)
    else:
        plotmesh3d.plotmesh(mesh=mesh, **kwargs)
    # if not 'ax' in kwargs or kwargs['ax']==plt: plt.show()

#=================================================================#
def meshWithBoundaries(meshdata, ax=plt):
    dim, meshdataismesh = _getDim(meshdata)
    kwargs = {}
    kwargs['ax'] = ax
    if dim==1:
        x, lines = meshdata.points[:, 0], meshdata.simplices
        plotmesh1d.meshWithBoundaries(x, lines, **kwargs)
    elif dim==2:
        if meshdataismesh:
            x, y, tris = meshdata.points[:,0], meshdata.points[:,1], meshdata.simplices
            kwargs['lines'] = meshdata.faces
            kwargs['bdrylabels'] = meshdata.bdrylabels
            if hasattr(meshdata, 'cell_labels'):
                kwargs['celllabels'] = meshdata.cell_labels
            if hasattr(meshdata, 'cellsoflabel'):
                kwargs['cellsoflabel'] = meshdata.cellsoflabel
        else:
            x, y, tris = meshdata[0], meshdata[1], meshdata[2]
            kwargs['lines'] = meshdata[3]
            kwargs['bdrylabels'] = meshdata[4]
        plotmesh2d.meshWithBoundaries(x, y, tris, **kwargs)
    else:
        if meshdataismesh:
            x, y, z, tets = meshdata.points[:,0], meshdata.points[:,1], meshdata.points[:,2], meshdata.simplices
            faces, bdrylabels = meshdata.faces, meshdata.bdrylabels
            plotmesh3d.meshWithBoundaries(x, y, z, tets, faces, bdrylabels, ax)
        else:
            plotmesh3d.meshWithBoundaries(meshdata, ax)
    if ax==plt: plt.show()

#=================================================================#
# def meshWithData(meshdata, point_data=None, cell_data=None, numbering=False, title=None, suptitle=None, addplots=[]):
def meshWithData(meshdata, **kwargs):

    dim, meshdataismesh = _getDim(meshdata)
    """
    meshdata    : either mesh or coordinates and connectivity
    point_data  : dictionary name->data
    cell_data  : dictionary name->data
    """
    newkwargs = kwargs.copy()

    if dim==1:
        if meshdataismesh:
            newkwargs['x'] = meshdata.points[:,0]
        else:
            newkwargs['x'] = meshdata[0]
        plotmesh1d.meshWithData(**newkwargs)
    elif dim==2:
        if meshdataismesh:
            newkwargs['x'] = meshdata.points[:,0]
            newkwargs['y'] = meshdata.points[:,1]
            newkwargs['tris'] = meshdata.simplices
            newkwargs['xc'] = meshdata.pointsc[:,0]
            newkwargs['yc'] = meshdata.pointsc[:,1]
        else:
            newkwargs['x'] = meshdata[0]
            newkwargs['y'] = meshdata[1]
            newkwargs['tris'] = meshdata[2]
            newkwargs['xc'] = meshdata.pointsc[3]
            newkwargs['yc'] = meshdata.pointsc[4]
        return plotmesh2d.meshWithData(**newkwargs)
    elif dim==3:
        if meshdataismesh:
            newkwargs['x'] = meshdata.points[:,0]
            newkwargs['y'] = meshdata.points[:,1]
            newkwargs['z'] = meshdata.points[:,2]
            newkwargs['tets'] = meshdata.simplices
            newkwargs['xc'] = meshdata.pointsc[:,0]
            newkwargs['yc'] = meshdata.pointsc[:,1]
            newkwargs['zc'] = meshdata.pointsc[:,2]
        else:
            newkwargs['x'] = meshdata[0]
            newkwargs['y'] = meshdata[1]
            newkwargs['z'] = meshdata[2]
            newkwargs['tets'] = meshdata[3]
            newkwargs['xc'] = meshdata.pointsc[4]
            newkwargs['yc'] = meshdata.pointsc[5]
            newkwargs['zc'] = meshdata.pointsc[6]
        return plotmesh3d.meshWithData(**newkwargs)
    else:
        raise ValueError(f"wrong dimension {dim=}")

#=================================================================#
def plotmeshWithNumbering(meshdata, **kwargs):
    if isinstance(meshdata,meshio._mesh.Mesh):
        types = [c.type for c in meshdata.cells]
        # print(f"{types=}")
        if 'tetra' in types: raise ValueError(f"so far only 2D")
        x, y  = meshdata.points[:,0], meshdata.points[:,1]
        for c, cb in meshdata.cells:
            if c=='triangle': tris = cb
            # elif c=='lines': faces = cb
    else:
        dim, meshdataismesh = _getDim(meshdata)
        if dim==3:
            raise NotImplementedError("3d not yet implemented")
        if meshdataismesh:
            x, y, tris, faces = meshdata.points[:,0], meshdata.points[:,1], meshdata.simplices, meshdata.faces
            kwargs['meshsides'] = faces
        else:
            x, y, tris = meshdata[0], meshdata[1], meshdata[2]

    if 'localnumbering' in kwargs and kwargs.pop('localnumbering'):
        fig, axs = plt.subplots(2, 3, figsize=(13.5, 8), squeeze=False)

        newkwargs = {}
        newkwargs['meshsides'] = faces
        newkwargs['cellsofsides'] = meshdata.cellsOfFaces
        newkwargs['sidesofcells'] = meshdata.facesOfCells
        newkwargs['meshnormals'] = meshdata.normals
        newkwargs['meshsigma'] = meshdata.sigma

        newkwargs['ax']= axs[0,0]
        plotmesh2d.mesh(x, y, tris, **newkwargs)

        newkwargs['ax']= axs[0,1]
        newkwargs['cellslocal']= True
        newkwargs['sides']= False
        plotmesh2d.mesh(x, y, tris, **newkwargs)

        newkwargs['ax']= axs[0,2]
        newkwargs['sideslocal']= True
        newkwargs['sides']= True
        newkwargs['cells']= False
        plotmesh2d.mesh(x, y, tris, **newkwargs)

        newkwargs['ax']= axs[1,0]
        newkwargs['nodes']= False
        newkwargs['sides']= False
        newkwargs['cells']= False
        newkwargs['cellsidelocal']= True
        plotmesh2d.mesh(x, y, tris, **newkwargs)

        newkwargs['ax']= axs[1,1]
        newkwargs['cellsidelocal']= False
        newkwargs['sidecelllocal']= True
        plotmesh2d.mesh(x, y, tris, **newkwargs)

        newkwargs['ax']= axs[1,2]
        newkwargs['normals']= True
        newkwargs['sidecelllocal']= False
        plotmesh2d.mesh(x, y, tris, **newkwargs)

    else:
        kwargs['ax']= plt
        plotmesh2d.mesh(x, y, tris, **kwargs)
    plt.show()

#=================================================================#
#=================================================================#
if __name__ == '__main__':
    import simplexmesh
    import geomdefs
    geometry = geomdefs.unitsquare.Unitsquare()
    mesh = simplexmesh.SimplexMesh(geometry=geometry, hmean=1)
    # plotmeshWithNumbering(mesh, localnumbering=True)
    plotmesh(mesh)



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  4 18:14:29 2016
@author: becker
"""
import numpy as np
import matplotlib.pyplot as plt
import pygmsh
from simfempy.tools.latexwriter import LatexWriter
import simfempy.meshes.pygmshext
#=================================================================#
class CompareMethods(object):
    """
    Run several times a list of methods (typically for comparison of different discretizations on a sequence of meshes)
    possible parameters:
      latex
      vtk
      plot
      plotpostprocs
      verb: in [0,5]
    """
    def __init__(self, methods, **kwargs):
        self.methods = methods
        self.dirname = "Results"
        self.latex = True
        self.plot = False
        self.plotpostprocs = False
        if 'clean' in kwargs and kwargs.pop("clean")==True:
            import os, shutil
            try: shutil.rmtree(os.getcwd() + os.sep + self.dirname)
            except: pass
        if 'latex' in kwargs: self.latex = kwargs.pop("latex")
        if 'plot' in kwargs: self.plot = kwargs.pop("plot")
        if 'plotpostprocs' in kwargs: self.plotpostprocs = kwargs.pop("plotpostprocs")
        if 'verbose' in kwargs:
            verbose = int(kwargs.pop("verbose"))
            if verbose > 0: self.latex = True
            if verbose > 1: self.plotpostprocs = True
            if verbose > 2: self.plot = True
        if 'h' in kwargs:
            self.h = kwargs.pop("h")
            self.paramname = kwargs.pop("paramname")
        else:
            self.paramname = "ncells"
            if 'niter' in kwargs: self.niter = kwargs.pop("niter")
            else: self.niter = -1
        self.parameters = []
    def _mesh_from_geom_or_fct(self, h):
        if hasattr(self,"createMesh"): return self.createMesh(h)
        if hasattr(pygmsh, "built_in"):
            mesh = pygmsh.generate_mesh(self.geom(h), verbose=False)
        else:
            with self.geom(h) as geom:
                mesh = geom.generate_mesh()
        return simfempy.meshes.simplexmesh.SimplexMesh(mesh=mesh)

    def compare(self, createMesh=None, geom=None, h=None, params=None):
        if createMesh is None: self.geom = geom
        else: self.createMesh = createMesh
        if self.paramname == "ncells":
            if h is None:
                mesh = self._mesh_from_geom_or_fct(self.h)
                gmshrefine = True
                if self.niter ==-1: raise KeyError("please give 'niter'")
                params = [mesh.ncells*mesh.dimension**i for i in range(self.niter)]
            else:
                params = h
                gmshrefine = False
        else:
            mesh = self._mesh_from_geom_or_fct(self.h)
            # mesh = SimplexMesh(geometry=geometry, hmean=self.h)
        for iter, param in enumerate(params):
            if self.paramname == "ncells":
                if gmshrefine:
                    mesh = simfempy.meshes.pygmshext.gmshRefine(mesh)
                else:
                    mesh = self._mesh_from_geom_or_fct(param)
                self.parameters.append(mesh.ncells)
            else:
                self.parameters.append(param)
            for name, method in self.methods.items():
                method.setMesh(mesh)
                self.dim = mesh.dimension
                if self.paramname != "ncells": method.setParameter(self.paramname, param)
                result = method.solve(iter, self.dirname)
                if self.plot:
                    from ..meshes import plotmesh
                    suptitle = "{}={}".format(self.paramname, self.parameters[-1])
                    plotmesh.meshWithData(mesh, data=result.data, title=name, suptitle=suptitle)
                    plt.show()
                resdict = result.info.copy()
                resdict.update(result.data['global'])
                # print(resdict)
                self.fillInfo(iter, name, resdict, len(params))
        if self.plotpostprocs:
            self.plotPostprocs(self.methods.keys(), self.paramname, self.parameters, self.infos)
        if self.latex:
            self.generateLatex(self.methods.keys(), self.paramname, self.parameters, self.infos)
        return  self.methods.keys(), self.paramname, self.parameters, self.infos
    def fillInfo(self, iter, name, info, n):
        if not hasattr(self, 'infos'):
            # first time - we have to generate some data
            self.infos = {}
            for key2, info2 in info.items():
                self.infos[key2] = {}
                for key3, info3 in info2.items():
                    self.infos[key2][key3] = {}
                    for name2 in self.methods.keys():
                        self.infos[key2][key3][name2] = np.zeros(shape=(n), dtype=type(info3))
        for key2, info2 in info.items():
            for key3, info3 in info2.items():
                self.infos[key2][key3][name][iter] = np.sum(info3)
    def generateLatex(self, names, paramname, parameters, infos):
        mesh = self.createMesh.__name__
        title = f"mesh({mesh})\\\\"
        for name, method in self.methods.items():
            title += f"{name}\\\\"
        title = title[:-2]
        # print("title = ", title)
        latexwriter = LatexWriter(dirname=self.dirname, title=title, author=self.__class__.__name__)
        for key, val in infos.items():
            kwargs = {'n': parameters, 'nname': paramname}
            if key == 'iter':
                newdict={}
                for key2, val2 in val.items():
                    for name in names:
                        newdict["{}-{}".format(key2, name)] = val2[name]
                kwargs['name'] = '{}'.format(key)
                kwargs['values'] = newdict
                kwargs['valformat'] = '3d'
                latexwriter.append(**kwargs)
            elif key == 'timer':
                for name in names:
                    newdict={}
                    for key2, val2 in val.items():
                        newdict["{}".format(key2)] = val2[name]
                    kwargs['name'] = '{}-{}'.format(key, name)
                    kwargs['values'] = newdict
                    kwargs['percentage'] = True
                    latexwriter.append(**kwargs)
            else:
                kwargs['redrate'] = (key=="error") and (paramname=="ncells")
                kwargs['diffandredrate'] = not kwargs['redrate'] and (paramname=="ncells")
                kwargs['dim'] = self.dim
                if key=="error" and len(self.methods)==1:
                    newdict={}
                    for key2, val2 in val.items():
                        newdict["{}".format(key2)] = val2[name]
                    kwargs['name'] = '{}-{}'.format(key, name)
                    kwargs['values'] = newdict
                    latexwriter.append(**kwargs)
                else:
                    for key2, val2 in val.items():
                        # print(f"{key=} {key2=}")
                        kwargs['name'] = '{}-{}'.format(key,key2)
                        kwargs['values'] = val2
                        latexwriter.append(**kwargs)
        latexwriter.write()
        latexwriter.compile()
    def computeOrder(self, ncells, values, dim):
        fnd = float(ncells[-1]) / float(ncells[0])
        order = -dim * np.log(values[-1] / values[0]) / np.log(fnd)
        return np.power(ncells, -order / dim), np.round(order,2)
    def plotPostprocs(self, names, paramname, parameters, infos):
        nmethods = len(names)
        self.reds = np.outer(np.linspace(0.2,0.8,nmethods),[0,1,1])
        self.reds[:,0] = 1.0
        self.greens = np.outer(np.linspace(0.2,0.8,nmethods),[1,0,1])
        self.greens[:,1] = 1.0
        self.blues = np.outer(np.linspace(0.2,0.8,nmethods),[1,1,0])
        self.blues[:,2] = 1.0
        singleplots = ['timer', 'iter']
        nplotsc = len(infos.keys())
        nplotsr = 0
        for key, val in infos.items():
            if key in singleplots: number=1
            else: number=len(val.keys())
            nplotsr = max(nplotsr, number)
        fig, axs = plt.subplots(nplotsr, nplotsc, figsize=(nplotsc * 3, nplotsr * 3), squeeze=False)
        cc = 0
        for key, val in infos.items():
            cr = 0
            for key2, val2 in val.items():
                for name in names:
                    if key == "error":
                        axs[cr,cc].loglog(parameters, val2[name], '-x', label="{}_{}".format(key2, name))
                        if self.paramname == "ncells":
                            orders, order = self.computeOrder(parameters, val2[name], self.dim)
                            axs[cr, cc].loglog(parameters, orders, '-', label="order {}".format(order))
                    else:
                        axs[cr, cc].plot(parameters, val2[name], '-x', label="{}_{}".format(key2, name))
                axs[cr, cc].legend()
                if key not in singleplots:
                    axs[cr, cc].set_title("{} {}".format(key, key2))
                    cr += 1
            if key in singleplots:
                axs[cr, cc].set_title("{}".format(key))
                cr += 1
            cc += 1
        plt.tight_layout()
        plt.show()
# ------------------------------------- #
if __name__ == '__main__':
    print("so far no test")


from . import applications, fems, meshes, solvers, tools

from .__about__ import __author__, __author_email__, __version__, __website__
__all__ = [
    "__version__",
    "__author__",
    "__author_email__",
    "__website__",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Dec  5 15:38:16 2016

@author: becker
"""

class StoppingData:
    def __init__(self):
        self.maxiter = 100
        self.atol = 1e-14
        self.rtol = 1e-10
        self.atoldx = 1e-14
        self.rtoldx = 1e-10
        self.divx = 1e8
        self.firststep = 1.0

        self.bt_maxiter = 50
        self.bt_omega = 0.5
        self.bt_c = 0.1



import numpy as np
import scipy.optimize
import time


# ----------------------------------------------------------------#
class RhsParam(object):
    def __init__(self, param):
        self.param = param

    def __call__(self, x, y, z):
        return self.param


# ----------------------------------------------------------------#
class Optimizer(object):

    def reset(self):
        self.r = None
        self.dr = None
        self.u = None
        self.z = None
        self.du = None

    def __init__(self, solver, **kwargs):
        self.solver = solver
        self.reset()
        self.fullhess = True
        self.gradtest = False
        self.hestest = False
        if 'fullhess' in kwargs: self.fullhess = kwargs.pop('fullhess')
        if 'gradtest' in kwargs: self.gradtest = kwargs.pop('gradtest')
        if 'hestest' in kwargs: self.hestest = kwargs.pop('hestest')
        if 'regularize' in kwargs:
            if not 'nmeasure' in kwargs: raise ValueError("If 'regularize' is given, we need 'nmeasure'")
            if not 'param0' in kwargs: raise ValueError("If 'regularize' is given, we need 'param0'")
            self.regularize = kwargs.pop('regularize')
            if self.regularize is not None:
                self.regularize = np.sqrt(self.regularize)
            else:
                self.regularize = 0
            self.param0 = kwargs.pop('param0')
        else:
            self.regularize = 0
        if 'nparam' in kwargs:
            self.nparam = kwargs.pop('nparam')
        else:
            self.nparam = solver.nparam
        if 'nmeasure' in kwargs:
            self.nmeasure = kwargs.pop('nmeasure')
        else:
            self.nmeasure = solver.nmeasure

        self.lsmethods = ['lm', 'trf', 'dogbox']
        self.minmethods = ['Newton-CG', 'trust-ncg', 'dogleg', 'trust-constr', 'SLSQP', 'BFGS', 'L-BFGS-B', 'TNC']
        self.methods = self.lsmethods + self.minmethods
        self.hesmethods = ['Newton-CG', 'trust-ncg', 'dogleg', 'trust-constr']
        self.boundmethods = ['trf', 'dogbox', 'L-BFGS-B', 'SLSQP', 'TNC']
        if not hasattr(solver, "computeM"):
            print("*** solver does not have 'computeM', setting 'fullhess=False'")
            self.fullhess = False

    # ------------------------------------------------------------------------
    def computeRes(self, param):
        self.r, self.u = self.solver.computeRes(param, self.u)
        if not hasattr(self, 'data0'):
            raise ValueError("Please set data0 !")
        self.r -= self.data0
        if self.regularize:
            self.r = np.append(self.r, self.regularize * (param - self.param0))
        return self.r

    # ------------------------------------------------------------------------
    def computeDRes(self, param):
        self.dr, self.du = self.solver.computeDRes(param, self.u, self.du)
        if self.regularize:
            self.dr = np.append(self.dr, self.regularize * np.eye(self.nparam), axis=0)
        return self.dr

    # ------------------------------------------------------------------------
    def computeJ(self, param):
        return 0.5 * np.linalg.norm(self.computeRes(param)) ** 2

    # ------------------------------------------------------------------------
    def testcomputeDRes(self, param, r, dr, u):
        eps = 1e-6
        if not dr.shape[0] == r.shape[0]:
            raise ValueError("wrong dimensions r.shape={} dr.shape={}".format(r.shape, dr.shape))
        for i in range(param.shape[0]):
            parameps = param.copy()
            parameps[i] += eps
            rp, up = self.solver.computeRes(parameps, u)
            parameps[i] -= 2 * eps
            rm, um = self.solver.computeRes(parameps, u)
            r2 = (rp - rm) / (2 * eps)
            if not np.allclose(dr[:self.nmeasure, i], r2):
                raise ValueError(
                    "problem in computeDRes:\ndr:\n{}\ndr(diff)\n{}\nparam={}\nrp={}\nrm={}".format(dr[:, i], r2, param,
                                                                                                    rp, rm))
            else:
                print(end='#')

    # ------------------------------------------------------------------------
    def computeDJ(self, param, computeRes=True):
        if self.r is None or computeRes:
            self.r = self.computeRes(param)
        self.dr, self.du = self.solver.computeDRes(param, self.u, self.du)
        # print("self.regularize", self.regularize)
        # print("self.r", self.r)
        # print("self.dr", self.dr.shape)
        if self.regularize:
            self.dr = np.append(self.dr, self.regularize * np.eye(self.nparam), axis=0)
        # print("self.dr", self.dr.shape)
        # print("grad=",np.dot(self.r, self.dr))
        if not self.gradtest:
            return np.dot(self.r, self.dr)
        self.testcomputeDRes(param, self.r, self.dr, self.u)

        if hasattr(self.solver, 'computeDResAdjW'):
            dr2 = self.solver.computeDResAdjW(param, self.u)
            if self.regularize: dr2 = np.append(dr2, self.regularize * np.eye(self.nparam), axis=0)
            if not np.allclose(self.dr, dr2):
                raise ValueError(
                    "dr('computeDRes') =\n{}\nbut dr2('computeDResAdjW') =\n{}".format(self.dr[:self.nmeasure],
                                                                                       dr2[:self.nmeasure]))
            # print("r", r.shape, "dr", dr.shape, "np.dot(r,dr)", np.dot(r,dr))

        if hasattr(self.solver, 'computeDResAdjW'):
            grad = np.dot(self.r, self.dr)
            grad2, self.z = self.solver.computeDResAdj(param, self.r[:self.nmeasure], self.u, self.z)
            if self.regularize:
                grad2 += self.regularize * self.regularize * (param - self.param0)
            if not np.allclose(grad, grad2):
                raise ValueError("different gradients\ndirect:  = {}\nadjoint: = {}".format(grad, grad2))
        return np.dot(self.r, self.dr)

    # ------------------------------------------------------------------------
    def testHessian(self, param, gn, M):
        H = gn + M
        H2 = np.zeros_like(H)
        eps = 1e-6
        for i in range(self.nparam):
            parameps = param.copy()
            parameps[i] += eps
            gradp = self.computeDJ(parameps, computeRes=True)
            parameps[i] -= 2 * eps
            gradm = self.computeDJ(parameps, computeRes=True)
            H2[i] = (gradp - gradm) / (2 * eps)
        if np.allclose(H, H2):
            print(end='@')
            return
        print("HD", H - H2)
        raise ValueError(
            "problem in testHessian:\nH:\n{}\nH(diff)\n{}\nparam={}\ngn={}\nM={}".format(H, H2, param, gn, M))

    # ------------------------------------------------------------------------
    def computeDDJ(self, param):
        if self.dr is None:
            self.dr, self.du = self.solver.computeDRes(param)
        gn = np.dot(self.dr.T, self.dr)
        if not self.fullhess:
            # print("Hes=", np.linalg.eigvals(gn))
            if self.hestest:
                self.testHessian(param, gn, np.zeros_like(gn))
            return gn

        self.z = self.solver.computeAdj(param, self.r[:self.nmeasure], self.u, self.z)
        M = self.solver.computeM(param, self.du, self.z)
        # print("gn", np.linalg.eigvals(gn), "M", np.linalg.eigvals(M))
        if not self.hestest:
            return gn + M
        self.testHessian(param, gn, M)
        return gn + M

    # ------------------------------------------------------------------------
    def create_data(self, refparam, percrandom=0, plot=False, printdata=False):
        nmeasures = self.nmeasure
        refdata, self.u = self.solver.computeRes(refparam)
        if refdata.reshape(-1).shape[0] != nmeasures:
            raise ValueError("wrong dim: nmeasures={} but refdata={}".format(nmeasures, refdata))
        if plot: self.solver.plot()
        perturbeddata = refdata * (1 + 0.5 * percrandom * (np.random.rand(nmeasures) - 2))
        if printdata:
            print("refparam", refparam)
            print("refdata", refdata)
            print("perturbeddata", perturbeddata)
        self.data0 = perturbeddata

    # ------------------------------------------------------------------------
    def minimize(self, x0, method, bounds=None, verbose=0, plot=False):
        self.reset()
        if not method in self.boundmethods: bounds = None
        # print("x0", x0, "method", method)
        hascost = True
        t0 = time.time()
        #         if bounds is None or method == 'lm': bounds = (-np.inf, np.inf)
        # scipy.optimize.show_options("minimize",method=method, disp=True)
        if method in self.lsmethods:
            if bounds is None: bounds = (-np.inf, np.inf)
            else: bounds = (bounds.lb, bounds.ub)
            print("bounds", bounds)
            info = scipy.optimize.least_squares(self.computeRes, jac=self.computeDRes, x0=x0,
                                                method=method, bounds=bounds, verbose=verbose)
        elif method in self.minmethods:
            hascost = False
            if method in self.hesmethods:
                hess = self.computeDDJ
            else:
                hess = None
            options = None
            # method = 'trust-constr'
            callback = None
            if method == 'Newton-CG':
                tol = None
                if verbose == 2:
                    options = {'xtol': 1e-16, 'disp':True}
                    def printiter(x0): print("x0", x0)
                    callback = printiter
            else:
                tol = 1e-12
            # if bounds and len(bounds) == 2: bounds = [bounds for l in range(len(x0))]
            info = scipy.optimize.minimize(self.computeJ, x0=x0, jac=self.computeDJ, hess=hess,
                                           method=method, bounds=bounds, tol=tol, callback=callback,
                                           options=options)
        else:
            raise NotImplementedError("unknown method '{}' known are {}".format(method, ','.join(self.methods)))
        dt = time.time() - t0
        # if method == 'trust-ncg': print(info)
        # print("info", info)
        if hascost:
            cost = info.cost
        else:
            cost = info.fun
        if hasattr(info, 'nhev'):
            nhev = info.nhev
        else:
            nhev = 0
        if hasattr(info, 'njev'):
            njev = info.njev
        else:
            njev = 0
        nfev = info.nfev
        if not info.success:
            print(10 * "@" + " no convergence!")
            nfev, njev, nhev = -1, -1, -1
        if hasattr(self.solver, 'param2x'):
            xf = self.solver.param2x(info.x)
        else:
            xf = info.x
        x = np.array2string(xf, formatter={'float_kind': lambda x: "%11.4e" % x})
        print(
            "{:^14s} x = {} J={:10.2e} nf={:4d} nj={:4d} nh={:4d} {:10.2f} s".format(method, x, cost, nfev, njev, nhev, dt))
        if plot:
            self.solver.plot(suptitle="{}".format(method))
        return xf, cost, info.nfev, njev, nhev, dt

    # ------------------------------------------------------------------------
    def testmethods(self, x0, methods, bounds=None, plot=False, verbose=0):
        values = {"J": [], "nf": [], "ng": [], "nh": [], "s": []}
        xall = np.empty(shape=(len(methods), self.nparam))
        valformat = {"J": "10.2e", "nf": "3d", "ng": "3d", "nh": "3d", "s": "6.1f"}
        for i, method in enumerate(methods):
            x, cost, nfev, njev, nhev, dt = self.minimize(x0=x0, method=method, bounds=bounds, plot=plot,
                                                          verbose=verbose)
            values["J"].append(cost)
            values["nf"].append(nfev)
            values["ng"].append(njev)
            values["nh"].append(nhev)
            values["s"].append(dt)
            xall[i] = x
        return values, valformat, xall
